using System;
using System.Collections;
using System.Collections.Generic;
using System.Text.RegularExpressions;

namespace Smdn.Applications.DirectoryAnalyzer.Analysis
{
  using FileSystem;

  public partial class Tree
  {
    private bool isScanned = false;
    private Tree parent = null;
    private Directory rootDir = null;
    private List<File> files = null;
    private List<Tree> subTrees = null;
    private uint depth = 0;

#region properties

    public bool IsScanned
    {
      get
      {
        return this.isScanned;
      }
    }

    public Tree Parent
    {
      get
      {
        return this.parent;
      }
    }

    public Directory RootDirectory
    {
      get
      {
        return this.rootDir;
      }
    }

    public File[] Files
    {
      get
      {
        return this.files.ToArray();
      }
    }

    public Tree[] SubTrees
    {
      get
      {
        return this.subTrees.ToArray();
      }
    }

    public uint Depth
    {
      get
      {
        return this.depth;
      }
    }

#endregion

#region construction

    public Tree(Directory rootDir)
    {
      if (null == rootDir) throw new ArgumentNullException("rootDir");

      this.parent = null;
      this.isScanned = false;
      this.rootDir = rootDir;
      this.subTrees = new List<Tree>();
      this.files = new List<File>();
    }

    private Tree(Tree parent, Directory rootDir)
      : this(rootDir)
    {
      this.parent = parent;
    }

#endregion

#region scanning

    /// <summary>
    /// ディレクトリをスキャンする
    /// </summary>
    /// <remarks>
    /// スキャン済みの場合は、現在保持しているツリー情報を破棄して再度スキャンする
    /// </remarks>
    /// <returns>ディレクトリの深さ</returns>
    public uint Scan(AnalysisContext.AnalysisStatus status, Regex regex, RegexPredicate<Directory> exclude)
    {
      if (this.isScanned)
      {
        this.subTrees = new List<Tree>();
        this.files = new List<File>();

        this.isScanned = false;
      }

      // ディレクトリ内を走査
      this.rootDir.ParseEntries(regex, exclude);

      foreach (Entry entry in this.rootDir)
      {
        if (entry is File)
        {
          // ファイルの場合
          this.files.Add(entry as File);

          status.AnalyzedEntryCount++;
        }
        else
        {
          // ディレクトリの場合、子ツリーを作成
          Tree subTree = new Tree(this, entry as Directory);

          // 子ツリーの追加のみ行い、後でまとめて走査する
          this.subTrees.Add(subTree);
        }
      }

      // 子ツリーを走査
      this.depth = 0;

      foreach (Tree subTree in this.subTrees)
      {
        uint d = subTree.Scan(status, regex, exclude);

        if (this.depth < d) this.depth = d;

        status.AnalyzedEntryCount++;
      }

      this.depth += 1;

      this.isScanned = true;

      return this.depth;
    }

#endregion

    public uint GetAllEntryCount()
    {
      uint count = 0;

      foreach (Tree tree in subTrees)
      {
        count += tree.GetAllEntryCount();
      }

      return count + GetEntryCount();
    }

    public uint GetEntryCount()
    {
      return (uint)files.Count + (uint)subTrees.Count;
    }

    public uint GetAllFileCount()
    {
      uint count = 0;

      foreach (Tree tree in subTrees)
      {
        count += tree.GetAllFileCount();
      }

      return count + (uint)files.Count;
    }

    public uint GetAllSubTreeCount()
    {
      uint count = 0;

      foreach (Tree tree in subTrees)
      {
        count += tree.GetAllSubTreeCount();
      }

      return count + (uint)subTrees.Count;
    }

    public Tree GetSubTreeOf(Directory subDirectory)
    {
      if (null == subDirectory) throw new ArgumentNullException("subDirectory");

      foreach (Tree subTree in subTrees)
      {
        if (subDirectory == subTree.RootDirectory)
        {
          return subTree;
        }
      }

      foreach (Tree subTree in subTrees)
      {
        Tree found = subTree.GetSubTreeOf(subDirectory);

        if (null != found) return found;
      }

      return null;
    }

    public Entry[] SearchAll(Regex regex, RegexPredicate<Entry> match)
    {
      if (null == regex) throw new ArgumentNullException("regex");

      List<Entry> matched = new List<Entry>();

      foreach (Tree tree in subTrees)
      {
        matched.AddRange(tree.SearchAll(regex, match));
      }

      matched.AddRange(Search(regex, match));

      return matched.ToArray();
    }

    public Entry[] Search(Regex regex, RegexPredicate<Entry> match)
    {
      if (null == regex) throw new ArgumentNullException("regex");
      if (null == match) throw new ArgumentNullException("match");

      List<Entry> matched = new List<Entry>();

      foreach (File file in files)
      {
        if (match(regex, file))
        {
          matched.Add(file);
        }
      }

      if (match(regex, rootDir))
      {
        matched.Add(rootDir);
      }

      return matched.ToArray();
    }

#region enumeration

    public IEnumerable GetSortedEnumerator(Comparison<Entry> comparer)
    {
      if (null == comparer) throw new ArgumentNullException("comparer");

      // エントリのリストを作る
      ArrayList list = new ArrayList();

      list.AddRange(this.subTrees);
      list.AddRange(this.files);

      object[] entries = list.ToArray();

      // ソートする
      Array.Sort(entries, delegate(object x, object y)
      {
        return comparer(
          (x is Tree) ? ((x as Tree).rootDir) : (x as Entry),
          (y is Tree) ? ((y as Tree).rootDir) : (y as Entry));
      });

      // 列挙子を返す
      foreach(object entry in entries)
      {
        yield return entry;
      }
    }

#endregion

#region dump

    public virtual void Dump()
    {
      Dump("");
    }

    protected virtual void Dump(string indent)
    {
      Console.WriteLine("{0} {1} {2}", indent, rootDir.Size, rootDir.FullName);

      indent += "  ";

      foreach (File file in files)
      {
        Console.WriteLine("{0} {1:P2} {2} {3}", indent, file.Size / (double)rootDir.Size, file.Size, file.Name);
      }

      foreach (Tree tree in subTrees)
      {
        tree.Dump(indent);
      }
    }

#endregion

    public override string ToString()
    {
      if (this.isScanned)
      {
        return string.Format("{0} (Files:{1} SubTrees:{2} Depth:{3})", this.rootDir, this.files.Count, this.subTrees.Count, this.depth);
      }
      else
      {
        return this.rootDir.ToString();
      }
    }
  }
}
