using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Reflection;
using System.Text;
using System.Text.RegularExpressions;

namespace Smdn.Applications.DirectoryAnalyzer
{
  using Analysis;
  using FileSystem;
  using Format;
  using Graphing;
  using Render;

  partial class DirectoryAnalyzer
  {
    private enum RunMode
    {
      Default,
      Visualize,
      Format,
      Usage,
    }

    private class ArgumentParser
    {
      public RunMode ParseArguments(string[] args, out Visualizer visualizer)
      {
        visualizer = null;

        // 必須オプションがない場合は、デフォルト動作
        if (args.Length < 1)
        {
          return RunMode.Default;
        }

#if LINUX
        // 使い方を表示
        if (args[0] == "--help")
        {
          return RunMode.Usage;
        }
#endif

        // 指定されたディレクトリが無い場合は使い方を表示して終了
        if (!System.IO.Directory.Exists(args[0]))
        {
          Console.Error.WriteLine("ディレクトリ{0}は存在しません。", args[0]);
          Console.Error.WriteLine();

          return RunMode.Usage;
        }

        // 解析されていない引数
        List<string> argsRemain = new List<string>(args);

        argsRemain.RemoveAt(0);

        // 引数を解析
        RunMode runMode = RunMode.Visualize;
        uint? outputWidth = null;
        uint? outputHeight = null;
        uint? graphingDepth = null;
        float? renderingInnerDiameter = null;
        Regex excludeRegex = null;
        FormatType? formatType = null;
        bool? silent = null;
        bool? compareEntryType = null;
        EntryCompareMode? entryCompareMode = null;
        Color? backgroundColor = null;
        Color? textColor = null;
        bool? preAnalyze = null;
        bool? drawSmallEntries = null;
        bool? drawNames = null;

        foreach (string a in args)
        {
          if (!a.StartsWith("/")) continue;

          // '/'削除、小文字化、値分割
          string arg = a.Substring(1).ToLower();
          string val = string.Empty;

          if (arg.Contains(":"))
          {
            val = a.Substring(arg.IndexOf(":") + 2);
            arg = arg.Substring(0, arg.IndexOf(":"));
          }

          switch (arg)
          {
            case "width": goto case "w";
            case "w":
              // 出力ファイルの幅
              outputWidth = uint.Parse(val);
              break;

            case "height": goto case "h";
            case "h":
              // 出力ファイルの高さ
              outputHeight = uint.Parse(val);
              break;

            case "depth": goto case "d";
            case "d":
              // 最大の深さ
              graphingDepth = uint.Parse(val);
              break;

            case "dia":
              // グラフの内径
              renderingInnerDiameter = float.Parse(val);
              if (renderingInnerDiameter < 0.1f || 0.9f < renderingInnerDiameter)
              {
                renderingInnerDiameter = null;
              }
              break;

            case "exclude":
              // 除外ディレクトリの正規表現
              excludeRegex = new Regex(val);
              break;

            case "xml":
              // XML形式で出力
              formatType = FormatType.Xml;
              break;

            case "csv":
              // CSV形式で出力
              formatType = FormatType.Csv;
              break;

            case "silent": goto case "s";
            case "quiet": goto case "s";
            case "s":
              // サイレント
              silent = true;
              break;

            case "individuate": goto case "i";
            case "i":
              // ファイルとディレクトリを区別
              compareEntryType = true;
              break;

            case "name": goto case "n";
            case "n":
              // ファイル名で並べ替え
              entryCompareMode = EntryCompareMode.ByName;
              break;

            case "bg":
              // 背景色
              backgroundColor = Color.FromKnownColor((KnownColor)Enum.Parse(typeof(KnownColor), val, true));
              break;

            case "text":
              // 文字色
              textColor = Color.FromKnownColor((KnownColor)Enum.Parse(typeof(KnownColor), val, true));
              break;

            case "detail":
              // 微少なエントリを描画する
              drawSmallEntries = true;
              break;

            case "nolabel":
              // エントリ名を描画しない
              drawNames = false;
              break;

            case "fast":
              // 事前解析しない
              preAnalyze = false;
              break;

            case "?": goto case "help";
            case "help":
              // 使い方を表示する
              return RunMode.Usage;

            default:
              continue;
          }

          // 解析済み
          argsRemain.Remove(a);
        }

        // 出力ファイル名
        string outputFile = null;

        if (argsRemain.Count == 0)
        {
          // 省略された場合は、カレントディレクトリに書き出す
          if (formatType.HasValue)
          {
            switch (formatType.Value)
            {
              case FormatType.Xml:
                outputFile = "graph.xml";
                break;

              case FormatType.Csv:
                outputFile = "graph.csv";
                break;
            }
          }
          else
          {
            outputFile = "graph.png";
          }
        }
        else
        {
          // 解析されなかった引数のうち、一番最後をファイル名と見なす
          outputFile = argsRemain[argsRemain.Count - 1];
        }

        // Visualizerのインスタンスを作成
        visualizer = new Visualizer(new Directory(args[0]));

        // AnalysisContextのパラメータを設定
        AnalysisContext.AnalysisParameter analysisParameter = visualizer.AnalysisContext.Parameter;

        analysisParameter.OutputFile = outputFile;
        if (graphingDepth.HasValue) analysisParameter.AnalysisDepth = (uint)graphingDepth;
        if (preAnalyze.HasValue) analysisParameter.PreAnalyze = (bool)preAnalyze;
        if (null != excludeRegex) analysisParameter.ExcludeRegex = excludeRegex;

        // GraphingContextのパラメータを設定
        GraphingContext.GraphingParameter graphingParameter = visualizer.GraphingContext.Parameter;

        if (graphingDepth.HasValue) graphingParameter.MaxDepth = (uint)graphingDepth;
        if (compareEntryType.HasValue || entryCompareMode.HasValue)
        {
          graphingParameter.Comparer = Comparison.GetComparer(entryCompareMode ?? EntryCompareMode.BySize, compareEntryType ?? false);
        }
        else
        {
          graphingParameter.Comparer = Comparison.BySizeIgnoreType;
        }
        //graphingParameter.Type = GraphType.Disc;

        // RenderingContextのパラメータを設定
        RenderingContext.RenderingParameter renderingParameter = visualizer.RenderingContext.Parameter;

        if (outputWidth.HasValue) renderingParameter.OutputWidth = (uint)outputWidth;
        if (outputHeight.HasValue) renderingParameter.OutputHeight = (uint)outputHeight;
        if (renderingInnerDiameter.HasValue) renderingParameter.InnerDiameterPercent = (float)renderingInnerDiameter;
        renderingParameter.DrawSmallEntries = drawSmallEntries ?? false;
        if (drawNames.HasValue) renderingParameter.DrawNames = (bool)drawNames;
        
        if (backgroundColor.HasValue) renderingParameter.BackgroundColor = (Color)backgroundColor;
        if (textColor.HasValue) renderingParameter.TextColor = (Color)textColor;

        // FormattingContextのパラメータを設定
        FormattingContext.FormattingParameter formattingParameter = visualizer.FormattingContext.Parameter;

        if (formatType.HasValue)
        {
          formattingParameter.Type = (FormatType)formatType;

          if (compareEntryType.HasValue || entryCompareMode.HasValue)
          {
            Comparison<Entry> comparer = Comparison.GetComparer(entryCompareMode ?? EntryCompareMode.BySize, compareEntryType ?? false);

            formattingParameter.Comparer = delegate(IGraphElement x, IGraphElement y)
            {
              return comparer(x.Entry, y.Entry);
            };
          }

          // 出力フォーマットが指定されている場合は、RunModeをFormatにする
          runMode = RunMode.Format;

          // 出力用のWriterを設定する
          switch (formatType)
          {
            case FormatType.Xml:
              formattingParameter.Writer = new StreamWriter(outputFile, false, Encoding.UTF8);
              break;

            case FormatType.Csv:
              formattingParameter.Writer = new StreamWriter(outputFile, false, Encoding.Default);
              break;
          }
        }

        // silentの場合は標準出力をTextWriter.Nullにリダイレクトする
        if (silent ?? false)
        {
          Console.SetOut(System.IO.TextWriter.Null);
        }

        return runMode;
      }
    }

    /// <summary>
    /// 使い方の表示
    /// </summary>
    private static void ShowUsage()
    {
      AssemblyFileVersionAttribute versionAttr = Attribute.GetCustomAttribute(Assembly.GetExecutingAssembly(), typeof(AssemblyFileVersionAttribute)) as AssemblyFileVersionAttribute;
      AssemblyCopyrightAttribute copyrightAttr = Attribute.GetCustomAttribute(Assembly.GetExecutingAssembly(), typeof(AssemblyCopyrightAttribute)) as AssemblyCopyrightAttribute;
      AssemblyDistributionPageAttribute distributionAttr = Attribute.GetCustomAttribute(Assembly.GetExecutingAssembly(), typeof(AssemblyDistributionPageAttribute)) as AssemblyDistributionPageAttribute;
      RenderingContext.RenderingParameter renderingParam = new RenderingContext.RenderingParameter();
#if LINUX
      string sampleDir = "/home/hoge/";
      string execCommand = "mono DirectoryAnalyzer.exe";
#elif WINDOWS
      string sampleDir = @"C:\Users\hoge\Documents\";
      string execCommand = "DirectoryAnalyzer.exe";
#endif

      string[] usage = new string[]
            {
                string.Format("DirectoryAnalyzer version {0}", versionAttr.Version),
                string.Format("{0}", copyrightAttr.Copyright),
                string.Format("{0}", distributionAttr.Url),
                "",
                "使い方:",
                "  DirectoryAnalyzer.exe ディレクトリ名 [オプション] [出力ファイル名]",
                "",
                "グラフのオプション:",
                "  次のオプションを組み合わせて使用できます。",
                "  /width:[幅]       出力する解析結果の画像の幅",
  string.Format("  /w:[幅]           デフォルト: {0}ピクセル", renderingParam.OutputWidth),
                "",
                "  /height:[高さ]    出力する解析結果の画像の高さ",
  string.Format("  /h:[高さ]         デフォルト: {0}ピクセル", renderingParam.OutputHeight),
                "",
                "  /depth:[深さ]     出力する解析結果の最大の深さ",
                "  /d:[深さ]         デフォルト: 指定なし(すべての階層を解析)",
                "",
                "  /dia:[内径]       出力するグラフの内径(0.1〜0.9)",
  string.Format("                    デフォルト: {0:F2}", renderingParam.InnerDiameterPercent),
                "",
                "  /bg:[KnownColor]  指定したKnownColorを背景色とする",
                "                    有効な値はSystem.Drawing.KnownColorで定義",
                "                    されている値",
                "",
                "  /text:[KnownColor]指定したKnownColorを文字色とする",
                "                    有効な値はSystem.Drawing.KnownColorで定義",
                "                    されている値",
                "",
                "  /detail           微少なエントリも描画する",
                "",
                "  /nolabel          グラフ内にエントリ名を描画しない",
                "",
                "出力形式に関するオプション:",
                "  /csv              CSV形式で結果を出力する",
                "",
                "  /xml              XML形式で結果を出力する",
                "",
                "  /individuate      ファイルとディレクトリを区別する",
                "  /i",
                "",
                "  /name             サイズ順ではなく名前順にソートする",
                "  /n",
                "",
                "動作に関するオプション:",
                "  /exclude:[Regex]  正規表現にマッチするディレクトリを除外して",
                "                    解析する",
                "",
                "  /fast             解析するディレクトリに含まれるエントリ数の",
                "                    計上を省略する",
                "",
                "  /silent           エラーを除く一切のメッセージを表示しない",
                "  /quiet",
                "  /s",
                "",
                "  /help, /?         使い方を表示する",
                "",
                "使用例:",
  string.Format("  {0} {1} analyzed.bmp", execCommand, sampleDir),
  string.Format("    {0}の解析結果をanalyzed.bmpに保存する", sampleDir),
                "",
  string.Format("  {0} {1} /bg:black /text:white /dia:0.2 analyzed.png", execCommand, sampleDir),
                "    背景を黒、文字色を白、ディスクの中心を小さめにして、",
                "    PNG形式で保存する",
                "",
  string.Format("  {0} /csv /i {1} files.csv", execCommand, sampleDir),
  string.Format("    サイズが大きいファイルの一覧をCSV形式で出力する", sampleDir),
                "",
            };

      foreach (string s in usage)
      {
        Console.Error.WriteLine(s);
      }
    }
  }
}
