using System;
using System.Drawing;
using System.IO;
using System.Threading;
#if LINUX || (!CONSOLE && WINDOWS)
using System.Windows.Forms;
#endif

namespace Smdn.Applications.DirectoryAnalyzer
{
  using Analysis;
  using FileSystem;
  using Graphing;
  using Render;

  public partial class DirectoryAnalyzer
  {
    [STAThread]
    public static void Main(string[] args)
    {
      (new DirectoryAnalyzer()).Run(args);
    }

    private delegate void Task();

    private void Run(string[] args)
    {
#if !CONSOLE && WINDOWS
#if !DEBUG
      // Windows-GUI版はログをファイルに書き出す
      TextWriter writer = null;
      try
      {
        writer = new StreamWriter(System.Reflection.Assembly.GetExecutingAssembly().GetName().Name + ".log");
        (writer as StreamWriter).AutoFlush = true;
        writer = TextWriter.Synchronized(writer);
      }
      catch
      {
      }
      finally
      {
        if (null != writer) Console.SetOut(writer);
      }
#endif
      Console.SetError(Console.Out);
#endif

      RunMode runmode = RunMode.Default;
      Visualizer visualizer = null;

#if CONSOLE
      ArgumentParser parser = new ArgumentParser();

      runmode = parser.ParseArguments(args, out visualizer);
#endif

      switch(runmode)
      {
        case RunMode.Usage:
          // 使い方
          ShowUsage();
          return;

        case RunMode.Default:
          // デフォルト動作
#if WINDOWS && CONSOLE
          ShowUsage();
          return;
#else
          Application.EnableVisualStyles();
          Application.SetCompatibleTextRenderingDefault(false);
          Application.Run(new UI.MainWindow());
          break;
#endif

        case RunMode.Format: goto case RunMode.Visualize;
        case RunMode.Visualize:
          // コマンドラインでの解析
          Console.WriteLine("Analysing directory: {0}", visualizer.AnalysisContext.Parameter.AnalysisDirectory.FullName);

          // ディレクトリを解析し、グラフを作成する
          Thread t = new Thread((ThreadStart)delegate
          {
            if (runmode == RunMode.Visualize)
            {
              visualizer.Visualize();
            }
            else
            {
              visualizer.Format();
            }
          });

          t.Start();

          for (; !t.Join(125); )
          {
            // 進捗を表示
            TaskContext.TaskStatus status = visualizer.Status;

            if (null != status)
            {
              ConsoleEx.RevertLocation();
              ConsoleEx.WriteProgress((int)status.Processed, (int)status.Total);
            }
          }

          // 描画/整形したグラフを保存する
          string outputFile = visualizer.AnalysisContext.Parameter.OutputFile;

          if (runmode == RunMode.Visualize)
          {
            visualizer.RenderingContext.Result.GraphImage.Save(outputFile);
          }
          else
          {
            visualizer.FormattingContext.Result.Formatted.Close();
          }

          Console.WriteLine("The graph was output to: {0}", outputFile);
          Console.WriteLine("Completed.");

          break;
      } // switch(runmode)
    }
  }
}
