using System;
using System.Collections.Generic;

#if LINUX
using DeviceInfo = Mono.Unix.UnixDriveInfo;
#elif WINDOWS
using DeviceInfo = System.IO.DriveInfo;
#endif

namespace Smdn.Applications.DirectoryAnalyzer.FileSystem
{
  public enum DeviceType
  {
    Fixed,
    CDRom,
    Network,
    Removable,
    Unknown,
  }

  public class Device
  {
    public static Device[] GetDevices()
    {
      List<Device> devices = new List<Device>();
#if LINUX
      foreach (DeviceInfo drive in DeviceInfo.GetDrives())
      {
        devices.Add(new Device(drive));
      }
#elif WINDOWS
      foreach (DeviceInfo drive in DeviceInfo.GetDrives())
      {
        devices.Add(new Device(drive));
      }
#endif

      return devices.ToArray();
    }

    private DeviceInfo info = null;

    private Device(DeviceInfo info)
    {
      if (null == info) throw new ArgumentNullException("info");
    
      this.info = info;
    }

    public DeviceInfo Info
    {
      get
      {
        return this.info;
      }
    }

    public Directory RootDirectory
    {
      get
      {
        return new Directory(this.info.RootDirectory.FullName);
      }
    }

    public DeviceType Type
    {
      get
      {
#if LINUX
        switch (this.info.DriveType)
        {
          case Mono.Unix.UnixDriveType.Fixed:
            return DeviceType.Fixed;

          case Mono.Unix.UnixDriveType.CDRom:
            return DeviceType.CDRom;

          case Mono.Unix.UnixDriveType.Network:
            return DeviceType.Network;

          case Mono.Unix.UnixDriveType.Removable:
            return DeviceType.Removable;

          default:
            return DeviceType.Unknown;
        }
#elif WINDOWS
        switch (this.info.DriveType)
        {
          case System.IO.DriveType.Fixed:
            return DeviceType.Fixed;

          case System.IO.DriveType.CDRom:
            return DeviceType.CDRom;

          case System.IO.DriveType.Network:
            return DeviceType.Network;

          case System.IO.DriveType.Removable:
            return DeviceType.Removable;

          default:
            return DeviceType.Unknown;
        }
#endif
      }
    }

    public string Name
    {
      get
      {
        return this.info.Name;
      }
    }

    public override string ToString()
    {
      string typeString = string.Empty;

      switch (this.Type)
      {
        case DeviceType.Fixed:
          typeString = "固定ドライブ";
          break;

        case DeviceType.CDRom:
          typeString = "光学ドライブ";
          break;

        case DeviceType.Network:
          typeString = "ネットワークドライブ";
          break;

        case DeviceType.Removable:
          typeString = "リムーバブルデバイス";
          break;

        default:
          typeString = string.Empty;
          break;
      }

      if (typeString == string.Empty)
      {
        return this.Name;
      }
      else
      {
        return string.Format("{0} ({1})", this.Name, typeString);
      }
    }
  }
}