using System;
using System.Collections;
using System.Collections.Generic;
using System.Text.RegularExpressions;

#if LINUX
using DirectoryInfo = Mono.Unix.UnixDirectoryInfo;
using FileInfo = Mono.Unix.UnixFileInfo;
using EntryInfo = Mono.Unix.UnixFileSystemInfo;
#elif WINDOWS
using DirectoryInfo = System.IO.DirectoryInfo;
using FileInfo = System.IO.FileInfo;
using EntryInfo = System.IO.FileSystemInfo;
#endif

namespace Smdn.Applications.DirectoryAnalyzer.FileSystem
{
  public partial class Directory : Entry, IEnumerable
  {
    private Entry[] subEntries = null;

    public Directory(DirectoryInfo directory)
      : base(directory)
    {
    }

    public Directory(string directoryPath)
      : base(new DirectoryInfo(directoryPath))
    {
    }

    public override string FullName
    {
      get { return (base.Info as DirectoryInfo).FullName; }
    }

    public override string Name
    {
      get { return (base.Info as DirectoryInfo).Name; }
    }

    private Size? cachedSize = null;

    public override Size Size
    {
      get
      {
        if (null == cachedSize)
        {
          cachedSize = Size.Zero;

          foreach (Entry entry in this)
          {
            cachedSize += entry.Size;
          }
        }

        return (Size)cachedSize;
      }
    }

    public IEnumerator GetEnumerator()
    {
      if (null == subEntries)
      {
        subEntries = ParseEntriesInternal(null, delegate(Regex regex, Directory directory)
        {
          return false;
        });
      }

      return subEntries.GetEnumerator();
    }

    /// <summary>
    /// 指定された正規表現にマッチするディレクトリを除いた上でディレクトリ内のエントリをパースする
    /// </summary>
    public void ParseEntries(Regex regex, RegexPredicate<Directory> exclude)
    {
      subEntries = ParseEntriesInternal(regex, exclude);
    }

    private Entry[] ParseEntriesInternal(Regex regex, RegexPredicate<Directory> exclude)
    {
      List<Entry> entries = new List<Entry>();

#if LINUX
      EntryInfo[] entryInfos = new EntryInfo[] { };

      try
      {
        entryInfos = (base.Info as DirectoryInfo).GetFileSystemEntries();
      }
      catch (UnauthorizedAccessException)
      {
        Console.Error.WriteLine("UnauthorizedAccessException in {0}", (base.Info as DirectoryInfo).FullName);
      }
      catch (System.IO.DirectoryNotFoundException)
      {
        Console.Error.WriteLine("DirectoryNotFoundException in {0}", (base.Info as DirectoryInfo).FullName);
      }
      catch (System.IO.IOException)
      {
        Console.Error.WriteLine("IOException in {0}", (base.Info as DirectoryInfo).FullName);
      }

      foreach (EntryInfo entryInfo in entryInfos)
      {
        if (entryInfo.IsRegularFile)
        {
          entries.Add(new File(entryInfo as FileInfo));
        }
        else if (entryInfo.IsDirectory)
        {
          Directory directory = new Directory(entryInfo as DirectoryInfo);

          // 正規表現にマッチする場合は除外する
          if (exclude(regex, directory)) continue;
          
          entries.Add(directory);
        }
        else
        {
          // TODO: シンボリックリンク等はいまのところ無視
          continue;
        }
      }
#elif WINDOWS
      EntryInfo[] entryInfos = new EntryInfo[] { };

      try
      {
        entryInfos = (base.Info as DirectoryInfo).GetFileSystemInfos();
      }
      catch (UnauthorizedAccessException)
      {
        Console.Error.WriteLine("UnauthorizedAccessException in {0}", (base.Info as DirectoryInfo).FullName);
      }
      catch (System.IO.DirectoryNotFoundException)
      {
        Console.Error.WriteLine("DirectoryNotFoundException in {0}", (base.Info as DirectoryInfo).FullName);
      }
      catch (System.IO.IOException)
      {
        Console.Error.WriteLine("IOException in {0}", (base.Info as DirectoryInfo).FullName);
      }

      foreach (EntryInfo entryInfo in entryInfos)
      {
        if (entryInfo is FileInfo)
        {
          entries.Add(new File(entryInfo as FileInfo));
        }
        else if (entryInfo is DirectoryInfo)
        {
          Directory directory = new Directory(entryInfo as DirectoryInfo);

          // 正規表現にマッチする場合は除外する
          if (exclude(regex, directory)) continue;

          entries.Add(directory);
        }
        else
        {
          // あり得ない?
          continue;
        }
      }
#endif

      return entries.ToArray();
    }

    public override bool Exists
    {
      get
      {
        return (base.Info as DirectoryInfo).Exists;
      }
    }

    public override Directory Parent
    {
      get
      {
        // UnixDirectoryInfoの親ディレクトリが見つからない場合、Parentはnullを返す
        // http://www.go-mono.com/docs/index.aspx?tlink=0@ecma%3a98%23UnixDirectoryInfo%2fP%2f1
        // DirectoryInfoがRootの場合、Parentはnullを返す
        // http://msdn2.microsoft.com/ja-jp/library/system.io.directoryinfo.parent(VS.80).aspx

        DirectoryInfo parent = (base.Info as DirectoryInfo).Parent;

        if (null == parent)
        {
          return null;
        }
        else
        {
          return new Directory(parent);
        }
      }
    }

#region entry count
    public class EntryCount
    {
      public uint FileCount = 0;
      public uint DirectoryCount = 0;
      public Size TotalSize = 0;

      public uint TotalEntryCount
      {
        get
        {
          return FileCount + DirectoryCount;
        }
      }
    }

    public static void GetEntryCount(Directory rootDir, bool recursive, ref EntryCount entryCount)
    {
      entryCount.FileCount = 0;
      entryCount.DirectoryCount = 0;
      entryCount.TotalSize = Size.Zero;

#if LINUX
      GetEntryCountWithReaddir(rootDir.FullName, recursive, ref entryCount);
#elif WINDOWS
      string dir = rootDir.FullName;

      // 末尾にはセパレータを付けない
      if (dir.EndsWith(System.IO.Path.DirectorySeparatorChar.ToString())) dir = dir.Substring(0, dir.Length - 1);

      // ネットワーク共有ではない場合、接頭辞を付ける
      if (!dir.StartsWith(@"\\")) dir = @"\\?\" + dir;

      GetEntryCountWithFind(dir, recursive, ref entryCount);
#endif
    }
#endregion
  }
}
