using System;
using System.Collections.Generic;
using System.Text.RegularExpressions;

namespace Smdn.Applications.DirectoryAnalyzer.Graphing.Disc
{
  using Analysis;
  using FileSystem;

  public class Disc : Graph
  {
    private TreePie rootPie = null;
    private uint depth = 0;

#region properties

    public uint Depth
    {
      get
      {
        return this.depth;
      }
    }

    public TreePie RootPie
    {
      get
      {
        return rootPie;
      }
    }

#endregion

#region construction

    public static Disc Create(GraphingContext context)
    {
      return new Disc(context);
    }

    private Disc(GraphingContext context)
      : base(context)
    {
      if (null == context) throw new ArgumentNullException("context");

      GraphingContext.GraphingParameter parameter = context.Parameter;

      this.depth = 0;

      this.rootPie = new TreePie(this, base.Tree, null);
      this.rootPie.Range = 1.0;
      this.rootPie.Volume = 0.0;

      uint maxDepth = (0 == parameter.MaxDepth) ? base.Tree.Depth : parameter.MaxDepth;

      if (base.Tree.Depth < maxDepth) maxDepth = base.Tree.Depth;

      GraphingContext.GraphingStatus status = context.Status;

      status.TotalEntryCount = base.Tree.GetAllEntryCount();
      status.GraphedEntryCount = 0;

      this.rootPie.SubPies = CreateSubPies(base.Tree, this.rootPie, maxDepth, parameter.Comparer, status, ref this.depth);

      status.GraphedEntryCount = status.TotalEntryCount;
    }

    private Pie[] CreateSubPies(Tree tree, TreePie parent, uint maxDepth, Comparison<Entry> comparer, GraphingContext.GraphingStatus status, ref uint depth)
    {
      // 深さが最大深度を超える場合は、解析しない
      if (maxDepth <= depth) return new Pie[] { };
      
      uint d = depth;

      // TODO: 今はサイズに基づいてRangeを決めるが、そのうち違う要素で
      List<Pie> pies = new List<Pie>();

      double totalSize = (double)tree.RootDirectory.Size;

      double parentRange = (totalSize == 0.0) ? 0.0 : parent.Range / totalSize;

      double argument = parent.Argument;

      // 各エントリについてパイを作成
      foreach (object entry in tree.GetSortedEnumerator(comparer))
      {
        Pie pie = null;

        if (entry is Tree)
        {
          // ツリーの場合、再帰的にパイを作成する
          Tree subTree = entry as Tree;

          pie = new TreePie(this, subTree, parent);

          pie.Argument = argument;
          pie.Range = parentRange * subTree.RootDirectory.Size;
          pie.Volume = 0.0;

          uint subDepth = d + 1;

          (pie as TreePie).SubPies = CreateSubPies(subTree, (pie as TreePie), maxDepth, comparer, status, ref subDepth);

          if (depth < subDepth) depth = subDepth;
        }
        else
        {
          // ファイルの場合
          pie = new Pie(this, entry as File, parent);

          pie.Argument = argument;
          pie.Range = parentRange * (entry as File).Size;
          pie.Volume = 0.0;
        }

        argument += pie.Range;

        pies.Add(pie);

        status.GraphedEntryCount++;
      }

      // ファイルが一つでもある場合は深さを1深く計上する
      if (d == depth && 0 < pies.Count) depth = d + 1;

      return pies.ToArray();
    }

#endregion

    public override IGraphElement[] GetAllElements()
    {
      return GetAllPies(rootPie);
    }

    private IGraphElement[] GetAllPies(TreePie pie)
    {
      List<IGraphElement> pies = new List<IGraphElement>();

      foreach (Pie subPie in pie.SubPies)
      {
        if (subPie is TreePie)
        {
          pies.AddRange(GetAllPies(subPie as TreePie));
        }
      }

      pies.AddRange(pie.SubPies);

      return pies.ToArray();
    }

    public override IGraphElement[] SearchAllElement(Regex regex, RegexPredicate<IGraphElement> match)
    {
      if (null == regex) throw new ArgumentNullException("regex");
      if (null == match) throw new ArgumentNullException("match");

      return SearchAllPie(rootPie, regex, match);
    }

    public override IGraphElement[] SearchElement(Regex regex, RegexPredicate<IGraphElement> match)
    {
      if (null == regex) throw new ArgumentNullException("regex");
      if (null == match) throw new ArgumentNullException("match");

      return SearchPie(rootPie, regex, match);
    }

    private IGraphElement[] SearchAllPie(TreePie pie, Regex regex, RegexPredicate<IGraphElement> match)
    {
      List<IGraphElement> matched = new List<IGraphElement>();

      foreach (Pie subPie in pie.SubPies)
      {
        if (match(regex, subPie))
        {
          matched.Add(subPie);
        }

        if (subPie is TreePie)
        {
          matched.AddRange(SearchAllPie(subPie as TreePie, regex, match));
        }
      }

      return matched.ToArray();
    }

    private IGraphElement[] SearchPie(TreePie pie, Regex regex, RegexPredicate<IGraphElement> match)
    {
      List<IGraphElement> matched = new List<IGraphElement>();

      foreach (Pie subPie in pie.SubPies)
      {
        if (match(regex, subPie))
        {
          matched.Add(subPie);
        }
      }

      return matched.ToArray();
    }
  }
}
