using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;

namespace Smdn.Applications.DirectoryAnalyzer.Render
{
  using Graphing;
  using Graphing.Disc;

  partial class Renderer
  {
    /// <summary>
    /// Disc用のRederer
    /// </summary>
    private class DiscRenderer : IDisposable
    {
      private Disc disc = null;
      private RenderingContext.RenderingParameter parameter = null;

      // テキストを描画する場合の、弧の最小の長さ
      private const double minTextArcLength = 16.0;

      // エントリを描画する場合の、最小の大きさ
      private const double minEntryRange = 0.001;

      // 寸法
      private uint maxDepth;
      private SizeF imageSize;
      private PointF discCenter;
      private float discOuterDiameter;
      private float discInnerDiameter;
      private float pieWidth;
      private RectangleF discBounds;
      private double pieAlphaDivider;

      // パイの枠線を描画するときのペン
      private Pen penPieBorder = null;

      // 背景を塗りつぶすときのブラシ
      private SolidBrush brushBackground = null;

      // テキストを描画するときのブラシ
      private SolidBrush brushText = null;

      // エントリ名を描画するときのフォント・フォーマット
      private Font fontName = null;
      private StringFormat stringFormatName = null;

      public DiscRenderer(RenderingContext.RenderingParameter parameter)
      {
        if (!(parameter.Graph is Disc)) throw new InvalidOperationException("Graph is not Disc");

        this.parameter = parameter;
        this.disc = parameter.Graph as Disc;

        CalculateMeasure();

        if (parameter.DrawNames)
        {
          fontName = new Font(SystemFonts.DefaultFont.FontFamily, 8.0f, FontStyle.Regular);

          stringFormatName = new StringFormat(StringFormat.GenericDefault);
          stringFormatName.Alignment = StringAlignment.Center;
          stringFormatName.LineAlignment = StringAlignment.Center;
          stringFormatName.Trimming = StringTrimming.EllipsisCharacter;
          stringFormatName.FormatFlags = StringFormatFlags.NoWrap;
        }
      }

      ~DiscRenderer()
      {
        (this as IDisposable).Dispose();
      }

      void IDisposable.Dispose()
      {
        if (null != fontName)
        {
          fontName.Dispose();
          fontName = null;
        }

        if (null != stringFormatName)
        {
          stringFormatName.Dispose();
          stringFormatName = null;
        }

        if (null != penPieBorder)
        {
          penPieBorder.Dispose();
          penPieBorder = null;
        }

        if (null != brushBackground)
        {
          brushBackground.Dispose();
          brushBackground = null;
        }

        if (null != brushText)
        {
          brushText.Dispose();
          brushText = null;
        }
      }

      /// <summary>
      /// 描画するためのペン・ブラシを準備する
      /// </summary>
      private void PrepareDrawingUtencils()
      {
        if (null == penPieBorder)
        {
          penPieBorder = new Pen(Color.FromArgb(0x60, Color.White), 1.0f);
        }

        if (null == brushBackground)
        {
          brushBackground = new SolidBrush(parameter.BackgroundColor);
        }

        if (null == brushText)
        {
          brushText = new SolidBrush(parameter.TextColor);
        }
      }

      /// <summary>
      /// ディスクの寸法を計算する
      /// </summary>
      private void CalculateMeasure()
      {
        // 描画する最大深度
        maxDepth = parameter.MaxDepth;

        if (disc.Depth < maxDepth) maxDepth = disc.Depth;
        if (0 == maxDepth) maxDepth = disc.Depth;

        imageSize = new SizeF((float)parameter.OutputWidth - 1.0f, (float)parameter.OutputHeight - 1.0f);
        discOuterDiameter = (imageSize.Width < imageSize.Height) ? imageSize.Width : imageSize.Height; // 外径
        discInnerDiameter = discOuterDiameter * parameter.InnerDiameterPercent; // 内径
        pieWidth = ((discOuterDiameter - discInnerDiameter) / 2.0f) / (float)maxDepth; // パイの幅(半径)
        discBounds = new RectangleF((imageSize.Width - discOuterDiameter) / 2.0f, (imageSize.Height - discOuterDiameter) / 2.0f, discOuterDiameter, discOuterDiameter); // 描画範囲
        discCenter = new PointF(imageSize.Width / 2.0f, imageSize.Height / 2.0f); // 円心
        pieAlphaDivider = 1.0 / (maxDepth + 1);// パイを描画するときのα値の除数
      }

      /// <summary>
      /// パイの外径の領域を計算する
      /// </summary>
      private RectangleF CalculatePieOuterDiameterBounds(uint depth)
      {
        float outerDiameter = discBounds.Width - pieWidth * 2.0f * (maxDepth - depth);

        return new RectangleF(discBounds.X + (discBounds.Width - outerDiameter) / 2.0f, discBounds.Y + (discBounds.Height - outerDiameter) / 2.0f, outerDiameter, outerDiameter);
      }

      /// <summary>
      /// Depth順に並べ替えたPieを取得する
      /// </summary>
      /// <remarks>
      /// <param name="minEntryRange">minEntryRange</param>で指定された値よりも小さいPieは含まない
      /// </remarks>
      private Pie[][] GetDepthSortedPies(Disc disc, uint maxDepth, double minEntryRange)
      {
        // 一番深いパイから描画するために、描画順にキューに入れる
        Queue<TreePie> enumeration = new Queue<TreePie>();
        List<List<Pie>> sorted = new List<List<Pie>>();

        sorted.Add(new List<Pie>());

        for (TreePie current = disc.RootPie; ; )
        {
          foreach (Pie pie in current.SubPies)
          {
            if (pie.Depth <= maxDepth)
            {
              if (sorted.Count == pie.Depth)
              {
                sorted.Add(new List<Pie>());
              }

              if (minEntryRange <= pie.Range)
              {
                sorted[(int)pie.Depth].Add(pie);
              }
            }

            if (pie is TreePie)
            {
              enumeration.Enqueue(pie as TreePie);
            }
          }

          if (0 == enumeration.Count)
          {
            break;
          }

          current = enumeration.Dequeue();
        }

        // 配列に変換
        Pie[][] pies = new Pie[sorted.Count][];

        for (int index = 0; index < sorted.Count; index++)
        {
          pies[index] = sorted[index].ToArray();
        }

        return pies;
      }

      /// <summary>
      /// 描画する
      /// </summary>
      public void Render(Graphics g, RenderingContext.RenderingStatus status)
      {
        // パイの色(描画色を最上層の親のパイと同じにする)
        Dictionary<Pie, Color> colorsForPie = new Dictionary<Pie, Color>();

        PrepareDrawingUtencils();

        // ディスクの外径を描画
        g.FillEllipse(brushBackground, discBounds);
        g.DrawEllipse(penPieBorder, discBounds);

        // パイを階層に分けて配列に格納する
        Pie[][] pies;

        if (parameter.DrawSmallEntries)
        {
          pies = GetDepthSortedPies(disc, maxDepth, 0.0);
        }
        else
        {
          // 微少なエントリを描画しない場合
          pies = GetDepthSortedPies(disc, maxDepth, minEntryRange);
        }

        // 描画対象のパイの数を計上
        status.RenderedEntryCount = 0;
        status.TotalEntryCount = 0;

        foreach (Pie[] p in pies)
        {
          status.TotalEntryCount += (uint)p.Length;
        }

        // 深い階層から順に描画していく
        for (uint depth = maxDepth; depth != 0; depth--)
        {
          // パイの外径
          RectangleF outerDiameterBounds = CalculatePieOuterDiameterBounds(depth);

          // 外径より内側を塗りつぶして一つ前の深さで描画した線を削除する
          if (depth != maxDepth) g.FillEllipse(brushBackground, outerDiameterBounds);

          foreach (Pie pie in pies[depth])
          {
            status.RenderedEntryCount++;

            // 描画する色を親のパイにあわせる
            Pie parentPie = pie.GetParent(1);
            Color colorPie;

            if (colorsForPie.ContainsKey(parentPie))
            {
              colorPie = colorsForPie[parentPie];
            }
            else
            {
              colorPie = CalculatePieColor(parentPie).ToColor();
            }

            // パイを描画する
            DrawPie(pie, false, parameter.DrawNames, g, colorPie, null, outerDiameterBounds);
          }
        }

        // 中心を塗りつぶす
        RectangleF discCenterBounds = new RectangleF(discCenter.X - discInnerDiameter / 2.0f, discCenter.Y - discInnerDiameter / 2.0f, discInnerDiameter, discInnerDiameter);

        g.FillEllipse(brushBackground, discCenterBounds);
        g.DrawEllipse(penPieBorder, discCenterBounds);

        // 統計情報を描画する
        if (parameter.DrawStatistics) DrawStatistics(g);
      }

      /// <summary>
      /// パイの色をHSI表色系で返す
      /// </summary>
      private Hsi CalculatePieColor(Pie pie)
      {
        return new Hsi((byte)Math.Truncate(255.0 * (pie.Argument + pie.Range * 0.5)), 128, 255);
      }

      /// <summary>
      /// パイを描画する
      /// </summary>
      private void DrawPie(Pie pie, bool drawInside, bool drawNames, Graphics g, Color colorPie, Pen penHighlightedBorder, RectangleF pieOuterDiameterBounds)
      {
        // 描画する弧の長さ
        double arcLength = Math.PI * pieOuterDiameterBounds.Width * pie.Range;

        // 深さに応じてα値を設定
        int alpha = (int)(31.0 + 224.0 * (maxDepth - pie.Depth + 1) * pieAlphaDivider);

        // パイを描画
        float argument = (float)pie.Argument * 360.0f - 90.0f;
        float sweep = (float)(pie.Range * 360.0f);

        using (Brush brushPie = new SolidBrush(Color.FromArgb(alpha, colorPie)))
        {
          g.FillPie(brushPie, pieOuterDiameterBounds.X, pieOuterDiameterBounds.Y, pieOuterDiameterBounds.Width, pieOuterDiameterBounds.Height, argument, sweep);
        }

        // 枠を描画
        if (null != penHighlightedBorder)
        {
          // ハイライト用のペンで描画する
          g.DrawPie(penHighlightedBorder, pieOuterDiameterBounds, argument, sweep);
        }
        else if (penPieBorder.Width <= arcLength)
        {
          // 描画する弧の長さがペンの幅より大きい場合のみ枠線を描画する
          g.DrawPie(penPieBorder, pieOuterDiameterBounds, argument, sweep);
        }

        // 内径の境界線を描画する場合
        if (drawInside)
        {
          RectangleF pieInnerDiameterBounds = RectangleF.FromLTRB(
            pieOuterDiameterBounds.Left + pieWidth,
            pieOuterDiameterBounds.Top + pieWidth,
            pieOuterDiameterBounds.Right - pieWidth,
            pieOuterDiameterBounds.Bottom - pieWidth);

          g.DrawPie(penPieBorder, pieInnerDiameterBounds, argument, sweep);
        }

        // 弧の長さが24px以上の場合は、名称を描画
        if (drawNames && minTextArcLength < arcLength)
        {
          DrawEntryName(pie, g, pieOuterDiameterBounds);
        }
      }

      /// <summary>
      /// パイにエントリ名を描画する
      /// </summary>
      private void DrawEntryName(Pie pie, Graphics g, RectangleF pieOuterDiameterBounds)
      {
        Matrix prevTransform = null;

        try
        {
          prevTransform = g.Transform;

          // パイの描画角にあわせて回転
          Matrix rotator = new Matrix();

          rotator.RotateAt((float)(pie.Argument + pie.Range * 0.5) * 360.0f, discCenter);

          g.Transform = rotator;

          // 描画する弧の長さ
          float arcLength = (float)(Math.PI * pieOuterDiameterBounds.Width * pie.Range);

          // テキストの描画領域
          RectangleF textBounds = new RectangleF(discCenter.X - arcLength / 2.0f, pieOuterDiameterBounds.Top, arcLength, pieWidth);

          // エントリ名
          string name = pie.Entry.Name;

          // ディレクトリの場合は末尾にセパレータを追加する
          if (pie.Entry is FileSystem.Directory) name += System.IO.Path.DirectorySeparatorChar;

          g.DrawString(name, fontName, brushText, textBounds, stringFormatName);
        }
        finally
        {
          if (null != prevTransform) g.Transform = prevTransform;
        }
      }

      /// <summary>
      /// ディスクの中心にルートディレクトリの統計情報を描画する
      /// </summary>
      private void DrawStatistics(Graphics g)
      {
        Analysis.Tree tree = disc.Tree;
        FileSystem.Directory rootDirectory = tree.RootDirectory;
        FileSystem.Size size = rootDirectory.Size;

        string text = string.Format("{0}\n\n{1}, {2}\n{3:N0} entries ({4:N0} files, {5:N0} directories)",
          rootDirectory.FullName,
          size.ToBinaryPrefixedString(),
          size.ToNumericString(),
          tree.GetAllEntryCount(),
          tree.GetAllFileCount(),
          tree.GetAllSubTreeCount());

        using (StringFormat format = new StringFormat(StringFormat.GenericDefault))
        {
          format.Alignment = StringAlignment.Center;
          format.LineAlignment = StringAlignment.Center;
          format.Trimming = StringTrimming.None;

          // 内径に内接する四角形
          float discInnerRadius = discInnerDiameter / 2.0f;
          RectangleF discInsideBounds = new RectangleF(discCenter.X - discInnerRadius * 0.707f, discCenter.Y - discInnerRadius * 0.707f, discInnerRadius * 1.414f, discInnerRadius * 1.414f);

          g.DrawString(text, SystemFonts.DefaultFont, brushText, discInsideBounds, format);
        }
      }

      /// <summary>
      /// Disc上の任意の位置に描画されるPieを取得する
      /// </summary>
      public Pie GetPieAt(PointF point)
      {
        // 点の位置から深さと範囲を算出する
        double dx = point.X - discCenter.X;
        double dy = point.Y - discCenter.Y;
        double rad = Math.Sqrt(dx * dx + dy * dy);

        uint depth = 0;

        if (rad * 2.0 < (double)discInnerDiameter)
        {
          // 中心からの距離の2倍が内径に満たない場合は、ディスクの内側
          return null;
        }
        else
        {
          depth = (uint)((rad * 2.0 - discInnerDiameter) / (pieWidth * 2.0)) + 1;
        }

        double range = 0.0;

        // Math.Atan2 = -π < θ ≦ π
        double arg = Math.Atan2(dy, dx) / (Math.PI * 2.0);

        if (arg < 0.0) arg += 1.0;

        range = arg - 0.75;

        if (range < 0.0)
        {
          range = range + 1.0;
        }

        // depth、rangeにマッチするパイを探す
        Queue<TreePie> enumeration = new Queue<TreePie>();
        List<List<Pie>> sorted = new List<List<Pie>>();

        sorted.Add(new List<Pie>());

        for (TreePie current = disc.RootPie; ; )
        {
          foreach (Pie pie in current.SubPies)
          {
            if (pie.Depth < depth)
            {
              if (sorted.Count == pie.Depth)
              {
                sorted.Add(new List<Pie>());
              }

              sorted[(int)pie.Depth].Add(pie);
            }
            else if (depth == pie.Depth)
            {
              arg = range - pie.Argument;

              if (0.0 <= arg && arg < pie.Range)
              {
                return pie;
              }
            }
            else
            {
              return null;
            }

            if (pie is TreePie)
            {
              enumeration.Enqueue(pie as TreePie);
            }
          }

          if (0 == enumeration.Count)
          {
            break;
          }

          current = enumeration.Dequeue();
        }

        return null;
      }

      /// <summary>
      /// Disc上の任意のPieを描画したBitmapを取得する
      /// </summary>
      public PartialGraphImage GetPieImage(IGraphElement[] elements, bool highlighted)
      {
        // elementsからPieのリストに変換する
        List<Pie> pies = new List<Pie>();

        foreach (IGraphElement element in elements)
        {
          pies.Add(element as Pie);
        }

        // 最深の深さを取得すると同時に、深さ別に分ける
        uint deepest = 0;
        Dictionary<uint, List<Pie>> rangedPies = new Dictionary<uint, List<Pie>>();

        pies.ForEach(delegate(Pie pie)
        {
          if (deepest < pie.Depth) deepest = pie.Depth;

          if (!rangedPies.ContainsKey(pie.Depth))
          {
            rangedPies[pie.Depth] = new List<Pie>();
          }

          rangedPies[pie.Depth].Add(pie);
        });

        // 寸法と最深を元に画像のサイズを決める
        int partialGraphSize = (int)(discInnerDiameter + (deepest - disc.RootPie.Depth) * pieWidth * 2.0f);

        // 戻り値のインスタンスを作成(大きさは余分にとる)
        PartialGraphImage partialGraph = new PartialGraphImage(elements, new Bitmap(partialGraphSize + 1, partialGraphSize + 1));

        // 描画するときのオフセットを求める
        partialGraph.ImageOffsetX = (int)(discBounds.X + (discBounds.Width - (float)partialGraphSize) / 2.0f);
        partialGraph.ImageOffsetY = (int)(discBounds.Y + (discBounds.Height - (float)partialGraphSize) / 2.0f);

        // 画像全体に描画するため、領域の中心をあわせる
        discCenter = new PointF(partialGraphSize / 2.0f, partialGraphSize / 2.0f);

        discBounds.X = discCenter.X - discBounds.Width / 2.0f;
        discBounds.Y = discCenter.Y - discBounds.Height / 2.0f;

        // 枠線を描画するためのペンを準備する
        PrepareDrawingUtencils();

        using (Graphics g = Graphics.FromImage(partialGraph.Image))
        {
          g.Clear(Color.Transparent);

          g.SmoothingMode = parameter.SmoothingMode;
          g.InterpolationMode = parameter.InterpolationMode;
          g.TextRenderingHint = parameter.TextRenderingHint;

          // 深い順に描画する
          for (uint depth = deepest; disc.RootPie.Depth != depth; depth--)
          {
            RectangleF outerDiameterBounds = CalculatePieOuterDiameterBounds(depth);

            // 内側を塗りつぶす
            g.FillEllipse(brushBackground, outerDiameterBounds);

            // 描画すべきパイがない場合は次の深さへ
            if (!rangedPies.ContainsKey(depth)) continue;

            foreach (Pie pie in rangedPies[depth])
            {
              Pie parentPie = pie.GetParent(1);

              Hsi hsi = CalculatePieColor(parentPie);

              if (highlighted)
              {
                using (Pen penHighlightedBorder = new Pen((new Hsi(hsi.Hue, 255, 192)).ToColor(), 4.0f))
                {
                  DrawPie(pie, true, parameter.DrawNames, g, hsi.ToColor(), penHighlightedBorder, outerDiameterBounds);
                }
              }
              else
              {
                DrawPie(pie, true, parameter.DrawNames, g, hsi.ToColor(), null, outerDiameterBounds);
              }
            }
          }

          // 中心を塗りつぶす
          RectangleF discCenterBounds = new RectangleF(discCenter.X - discInnerDiameter / 2.0f, discCenter.Y - discInnerDiameter / 2.0f, discInnerDiameter, discInnerDiameter);

          g.FillEllipse(brushBackground, discCenterBounds);
        }

        // 塗りつぶした箇所を透過させる
        partialGraph.Image.MakeTransparent(parameter.BackgroundColor);

        return partialGraph;
      }
    } // class DiscRenderer
  } // class Renderer
}