using System;
using System.Drawing;
using System.Text.RegularExpressions;
using System.Windows.Forms;

namespace Smdn.Applications.DirectoryAnalyzer.UI
{
  using Analysis;
  using FileSystem;
  using Graphing;
  using Util;

  public partial class MainWindow
  {
    private GraphElementContextMenuStrip directoryContextMemuStrip = null;
    private GraphElementContextMenuStrip fileContextMemuStrip = null;

    private struct EventItemDesc
    {
      public string Text;
      public EventHandler Event;

      public EventItemDesc(string text, EventHandler @event)
      {
        this.Text = text;
        this.Event = @event;
      }
    }

    private void InitializeHandler()
    {
      // 保存
      toolStripMenuItemFileSave_Graph.Click += delegate
      {
        SaveResult(FormatType.Graph);
      };

      toolStripMenuItemFileSave_Csv.Click += delegate
      {
        SaveResult(FormatType.Csv);
      };

      toolStripMenuItemFileSave_Xml.Click += delegate
      {
        SaveResult(FormatType.Xml);
      };
      
      // 終了
      toolStripMenuItemFile_Quit.Click += delegate
      {
        this.Close();
      };

      // グラフに関するオプション
      toolStripMenuItemOption_Graph.DropDownOpening += delegate
      {
        toolStripMenuItemOptionGraph_DrawSmallEntries.Checked = viewer.DrawSmallEntries;
        toolStripMenuItemOptionGraph_DrawHighQuality.Checked = viewer.DrawHighQuality;

        if (this.entryCompareMode == EntryCompareMode.BySize)
        {
          toolStripMenuItemOptionGraph_CompareBySize.Checked = true;
          toolStripMenuItemOptionGraph_CompareByName.Checked = false;
        }
        else if (this.entryCompareMode == EntryCompareMode.ByName)
        {
          toolStripMenuItemOptionGraph_CompareBySize.Checked = false;
          toolStripMenuItemOptionGraph_CompareByName.Checked = true;
        }

        toolStripMenuItemOptionGraph_CompareEntryType.Checked = this.compareEntryType;
      };

      toolStripMenuItemOptionGraph_DrawSmallEntries.Click += delegate
      {
        SetViewerRenderingOption(!viewer.DrawSmallEntries, viewer.DrawHighQuality, viewer.BackgroundColor, viewer.TextColor);
      };

      toolStripMenuItemOptionGraph_DrawHighQuality.Click += delegate
      {
        SetViewerRenderingOption(viewer.DrawSmallEntries, !viewer.DrawHighQuality, viewer.BackgroundColor, viewer.TextColor);
      };

      toolStripMenuItemOptionGraph_CompareBySize.Click += delegate
      {
        SetViewerGraphingOption(EntryCompareMode.BySize, this.compareEntryType);
      };

      toolStripMenuItemOptionGraph_CompareByName.Click += delegate
      {
        SetViewerGraphingOption(EntryCompareMode.ByName, this.compareEntryType);
      };

      toolStripMenuItemOptionGraph_CompareEntryType.Click += delegate
      {
        SetViewerGraphingOption(this.entryCompareMode, !this.compareEntryType);
      };

      toolStripMenuItemOptionGraph_BackgroundColor.DropDownOpening += delegate
      {
        if (viewer.BackgroundColor.Equals(Color.White))
        {
          toolStripMenuItemOptionGraphBackgroundColor_White.Checked = true;
          toolStripMenuItemOptionGraphBackgroundColor_Black.Checked = false;
        }
        else
        {
          toolStripMenuItemOptionGraphBackgroundColor_White.Checked = false;
          toolStripMenuItemOptionGraphBackgroundColor_Black.Checked = true;
        }
      };

      toolStripMenuItemOptionGraphBackgroundColor_White.Click += delegate
      {
        SetViewerRenderingOption(viewer.DrawSmallEntries, viewer.DrawHighQuality, Color.White, Color.Black);
      };

      toolStripMenuItemOptionGraphBackgroundColor_Black.Click += delegate
      {
        SetViewerRenderingOption(viewer.DrawSmallEntries, viewer.DrawHighQuality, Color.Black, Color.White);
      };

      // 検索結果に関するオプション
      toolStripMenuItemOption_Tabulation.DropDownOpening += delegate
      {
        foreach (ToolStripItem item in toolStripMenuItemOption_Tabulation.DropDownItems)
        {
          if (item is ToolStripMenuItem)
          {
            (item as ToolStripMenuItem).Checked = false;
          }
        }

        switch (tabulator.TabulateMax)
        {
          case 10:
            toolStripMenuItemOptionTabulation_Upto10.Checked = true;
            break;
          
          case 20:
            toolStripMenuItemOptionTabulation_Upto20.Checked = true;
            break;
          
          case 30:
            toolStripMenuItemOptionTabulation_Upto30.Checked = true;
            break;
          
          case 50:
            toolStripMenuItemOptionTabulation_Upto50.Checked = true;
            break;

          case 100:
            toolStripMenuItemOptionTabulation_Upto100.Checked = true;
            break;

          case 0:
            toolStripMenuItemOptionTabulation_All.Checked = true;
            break;

          default:
            toolStripMenuItemOptionTabulation_Upto30.Checked = true;
            break;
        }
      };

      foreach (object[] entry in new object[][] 
        {
          new object[] { toolStripMenuItemOptionTabulation_Upto10, 10 },
          new object[] { toolStripMenuItemOptionTabulation_Upto20, 20 },
          new object[] { toolStripMenuItemOptionTabulation_Upto30, 30 },
          new object[] { toolStripMenuItemOptionTabulation_Upto50, 50 },
          new object[] { toolStripMenuItemOptionTabulation_Upto100, 100},
          new object[] { toolStripMenuItemOptionTabulation_All, 0 },
        })
      {
        (entry[0] as ToolStripMenuItem).Tag = entry[1];
        (entry[0] as ToolStripMenuItem).Click += delegate(object sender, EventArgs e)
        {
          SetTabulateMax((int)(sender as ToolStripMenuItem).Tag);
        };
      }

      // フォルダを選択して解析
      EventHandler visualizeSelectedDirectoryHandler = delegate
      {
        Visualize(SelectVisualizeDirectory());
      };

      toolStripMenuItemVisualize_Directory.Click += visualizeSelectedDirectoryHandler;
      toolStripButtonVisualizeDirectory.Click += visualizeSelectedDirectoryHandler;

      // 中断
      EventHandler abortVisualizeHandler = delegate
      {
        AbortVisualize();
      };

      toolStripMenuItemVisualize_Abort.Click += abortVisualizeHandler;
      toolStripButtonAbortVisualize.Click += abortVisualizeHandler;

      // SpecialFolderを解析
      EventHandler visualizeDirectoryHandler = delegate(object sender, EventArgs e)
      {
        Visualize((sender as IDirectorySelectionControl).Directory);
      };

#if LINUX
      object[][] specialFolders = new object[][] {
        new object[] {Environment.SpecialFolder.Desktop, "デスクトップ"},
        new object[] {Environment.SpecialFolder.MyDocuments,"ホーム"},
      };
#elif WINDOWS
      object[][] specialFolders = new object[][] {
        new object[] {Environment.SpecialFolder.Desktop, "デスクトップ"},
        new object[] {Environment.SpecialFolder.MyDocuments,"マイ ドキュメント"},
        new object[] {Environment.SpecialFolder.ProgramFiles, "プログラム ファイル"},
      };
#endif

      foreach (object[] specialFolder in specialFolders)
      {
        string path = Environment.GetFolderPath((Environment.SpecialFolder)specialFolder[0]);

        if (string.IsNullOrEmpty(path)) continue;

        Directory directory = new Directory(path);

        DirectorySelectionToolStripMenuItem item = new DirectorySelectionToolStripMenuItem(directory);

        item.Text = specialFolder[1] as string;
        item.Click += visualizeDirectoryHandler;

        toolStripMenuItemVisualize.DropDownItems.Add(item);

        DirectorySelectionToolStripButton button = new DirectorySelectionToolStripButton(directory);

        button.Text = specialFolder[1] as string;
        button.ToolTipText = button.Text;
        button.DisplayStyle = ToolStripItemDisplayStyle.Image;
        button.Click += visualizeDirectoryHandler;

        toolStripMain.Items.Add(button);
      }

      toolStripMenuItemVisualize.DropDownItems.Add(new ToolStripSeparator());

      toolStripMain.Items.Add(new ToolStripSeparator());

      // デバイスを解析
      EventHandler visualizeDeviceHandler = delegate(object sender, EventArgs e)
      {
        Visualize((sender as IDeviceSelectionControl).Device);
      };

      foreach (Device device in Device.GetDevices())
      {
        DeviceSelectionToolStripMenuItem item = new DeviceSelectionToolStripMenuItem(device);

        string text;
#if LINUX
        text = device.ToString();

        // ニーモニック(/xxx/...の形式になるので、2文字目をニーモニック文字とする)
        if (2 <= text.Length && Char.IsLetter(text, 1))
        {
          text = text.Substring(0, 1) + "&" + text.Substring(1);
        }
#elif WINDOWS
        text = device.ToString();

        // ニーモニック(X:\...の形式になるので、1文字目をニーモニック文字とする)
        if (Char.IsLetter(text, 0))
        {
          text = "&" + text;
        }
#endif
        item.Text = text;
        item.Click += visualizeDeviceHandler;

        toolStripMenuItemVisualize.DropDownItems.Add(item);

        DeviceSelectionToolStripButton button = new DeviceSelectionToolStripButton(device);

        button.Text = string.Format("{0}", device.ToString());
        button.ToolTipText = button.Text;
        button.DisplayStyle = ToolStripItemDisplayStyle.Image;
        button.Click += visualizeDeviceHandler;

        toolStripMain.Items.Add(button);
      }

      // 後ろに履歴を追加するためのセパレータ
      toolStripMain.Items.Add(new ToolStripSeparator());

      // 一つ上のフォルダへ
      toolStripButtonAscendDirectory.Click += delegate
      {
        AscendDirectory();
      };

      // 戻る・進む
      toolStripSplitButtonBackHistory.ButtonClick += delegate
      {
        BackHistory(1);
      };

      toolStripSplitButtonForwardHistory.ButtonClick += delegate
      {
        ForwardHistory(1);
      };

      // About
      toolStripMenuItemHelp_About.Click += delegate
      {
        (new AboutDialog()).ShowDialog(this);
      };

      // 検索
      textBoxSearch.QueryEntered += delegate(object sender, EventArgs e)
      {
        if (radioButtonSearchRegex.Checked)
        {
          SearchByRegex((sender as QueryTextBox).Text);
        }
        else if (radioButtonSearchWildCard.Checked)
        {
          SearchByWildcard((sender as QueryTextBox).Text);
        }
      };

      // Viewer関連
      viewer.ElementAtCursorChanged += delegate(object sender, ViewerMouseEventArgs e)
      {
        Viewer v = (sender as Viewer);

        if (null == e.Element)
        {
          toolTipElementAtCursor.SetToolTip(v.ViewingControl, string.Empty);

          SetStatistics(v.Graph.Tree);
        }
        else
        {
          if (e.Element.Entry is Directory)
          {
            toolTipElementAtCursor.SetToolTip(v.ViewingControl, e.Element.Entry.Name + System.IO.Path.DirectorySeparatorChar);
          }
          else
          {
            toolTipElementAtCursor.SetToolTip(v.ViewingControl, e.Element.Entry.Name);
          }

          SetStatistics(e.Element);
        }
      };

      viewer.ElementClicked += delegate(object sender, ViewerMouseEventArgs e)
      {
        if (MouseButtons.Left == (e.Button & MouseButtons.Left))
        {
          // ディレクトリがクリックされた場合は、そのディレクトリを見る
          if (e.Element is ITreeGraphElement)
          {
            ViewDirectory(e.Element as ITreeGraphElement);
          }
        }
        else if (MouseButtons.Right == (e.Button & MouseButtons.Right))
        {
          if (e.Element is ITreeGraphElement)
          {
            directoryContextMemuStrip.Show((sender as Viewer).ViewingControl, e.X, e.Y, e.Element);
          }
          else
          {
            fileContextMemuStrip.Show((sender as Viewer).ViewingControl, e.X, e.Y, e.Element);
          }
        }
      };

      viewer.StatusChanged += delegate(object sender, ViewerStatusEventArgs e)
      {
        SetFinalStatus(e.Status);
      };

      viewer.HistoryChanged += delegate
      {
        HistoryChanged();
      };

      viewer.GraphRootChanged += delegate
      {
        GraphRootChanged();
      };

      // Tabulator関連
      tabulator.ItemClicked += delegate(object sender, TabulatorMouseEventArgs e)
      {
        if (MouseButtons.Left == (e.Button & MouseButtons.Left))
        {
          // do nothing
        }
        else if (MouseButtons.Right == (e.Button & MouseButtons.Right))
        {
          if (e.Element is ITreeGraphElement)
          {
            directoryContextMemuStrip.Show((sender as Tabulator).TabulatingListView, e.X, e.Y, e.Element);
          }
          else
          {
            fileContextMemuStrip.Show((sender as Tabulator).TabulatingListView, e.X, e.Y, e.Element);
          }
        }
      };

      tabulator.SelectedElementChanged += delegate
      {
        SelectedElementChanged();
      };

      // Viewer・Tabulatorを右クリックしたときに表示されるコンテキストメニュー
      fileContextMemuStrip = new GraphElementContextMenuStrip();
      directoryContextMemuStrip = new GraphElementContextMenuStrip();

      for (int i = 0; i < 2; i++)
      {
        ContextMenuStrip strip = null;
        EventItemDesc[] descs = null;

        switch (i)
        {
          case 0:
            strip = fileContextMemuStrip;
            descs = new EventItemDesc[]
            {
              new EventItemDesc("このファイルのあるフォルダを見る", delegate(object sender, EventArgs e)
              {
                ViewDirectory(((sender as ToolStripMenuItem).Owner as GraphElementContextMenuStrip).Element.Parent);
              }),

              new EventItemDesc("-", null),

              new EventItemDesc("ファイルを開く", delegate(object sender, EventArgs e)
              {
                FileManager.OpenEntry(((sender as ToolStripMenuItem).Owner as GraphElementContextMenuStrip).Element.Entry);
              }),

              new EventItemDesc("このファイルのあるフォルダを開く", delegate(object sender, EventArgs e)
              {
                FileManager.OpenParentEntry(((sender as ToolStripMenuItem).Owner as GraphElementContextMenuStrip).Element.Entry);
              }),

              new EventItemDesc("-", null),

              new EventItemDesc("プロパティ", delegate(object sender, EventArgs e)
              {
                FileManager.ShowProperty(((sender as ToolStripMenuItem).Owner as GraphElementContextMenuStrip).Element.Entry);
              }),
            };
            break;

          case 1:
            strip = directoryContextMemuStrip;
            descs = new EventItemDesc[]
            {
              new EventItemDesc("フォルダを見る", delegate(object sender, EventArgs e)
              {
                ViewDirectory(((sender as ToolStripMenuItem).Owner as GraphElementContextMenuStrip).Element as ITreeGraphElement);
              }),

              new EventItemDesc("このフォルダのあるフォルダを見る", delegate(object sender, EventArgs e)
              {
                ViewDirectory(((sender as ToolStripMenuItem).Owner as GraphElementContextMenuStrip).Element.Parent);
              }),

              new EventItemDesc("-", null),

              new EventItemDesc("フォルダを開く", delegate(object sender, EventArgs e)
              {
                FileManager.OpenEntry(((sender as ToolStripMenuItem).Owner as GraphElementContextMenuStrip).Element.Entry);
              }),

              new EventItemDesc("このフォルダのあるフォルダを開く", delegate(object sender, EventArgs e)
              {
                FileManager.OpenParentEntry(((sender as ToolStripMenuItem).Owner as GraphElementContextMenuStrip).Element.Entry);
              }),

              new EventItemDesc("-", null),

              new EventItemDesc("プロパティ", delegate(object sender, EventArgs e)
              {
                FileManager.ShowProperty(((sender as ToolStripMenuItem).Owner as GraphElementContextMenuStrip).Element.Entry);
              }),
            };
            break;
        }

        foreach (EventItemDesc desc in descs)
        {
          ToolStripItem item = null;

          if ("-" == desc.Text)
          {
            item = new ToolStripSeparator();
          }
          else
          {
            item = new ToolStripMenuItem(desc.Text);
          }

          if( null != desc.Event)
          {
            item.Click += desc.Event;
          }

          strip.Items.Add(item);
        }
      }
    }
  }
}
