using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.IO;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;
using System.Windows.Forms;

namespace Smdn.Applications.DirectoryAnalyzer.UI
{
  using Analysis;
  using FileSystem;
  using Format;
  using Graphing;
  using Render;

  public partial class MainWindow : Form
  {
    private enum FormatType
    {
      Graph,
      Xml,
      Csv,
    }

    private const int visualizationHistoryMax = 5;

    private Viewer viewer = null;
    private Tabulator tabulator = null;

    private EntryCompareMode entryCompareMode = EntryCompareMode.BySize;
    private bool compareEntryType = true;

    /// <summary>
    /// コンストラクタ
    /// </summary>
    public MainWindow()
    {
      InitializeComponent();

      Initialize();
    }

    private void Initialize()
    {
      viewer = new Viewer(this.pictureBoxViewer);
      tabulator = new Tabulator(this.listViewTabulator);

      // ハンドラの割り当て
      InitializeHandler();

      // アイコンの割り当て
      IntializeIcon();

      this.Text = string.Format("{0} ver {1}", Application.ProductName, Application.ProductVersion);
      this.StartPosition = FormStartPosition.CenterScreen;

      // 起動直後は隠しておく
      toolStripProgressBarStatus.Visible = false;

      toolStripStatusLabelStatus.Text = "解析するフォルダまたはドライブを選択してください";

      toolStripSplitButtonForwardHistory.Enabled = false;
      toolStripSplitButtonBackHistory.Enabled = false;
    }

    #region "visualization"

    private Thread visualizingThread = null;
    private ManualResetEvent visualizingEvent = new ManualResetEvent(false);
    private System.Threading.Timer visualizingTimer = null;

    private Directory SelectVisualizeDirectory()
    {
      FolderBrowserDialog dialog = new FolderBrowserDialog();

      dialog.Description = "解析するフォルダを選択してください。";
      dialog.RootFolder = Environment.SpecialFolder.Desktop;
      dialog.ShowNewFolderButton = false;
      
      DialogResult result = dialog.ShowDialog();

      if (DialogResult.OK == result)
      {
        try
        {
          return new Directory(dialog.SelectedPath);
        }
        catch (NotSupportedException)
        {
          MessageBox.Show(this, "そのフォルダは解析できません。", Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Exclamation);

          return null;
        }
      }
      else
      {
        return null;
      }
    }

    private void Visualize(Device device)
    {
      switch (device.Type)
      {
        case DeviceType.CDRom: goto case DeviceType.Network;
        case DeviceType.Removable: goto case DeviceType.Network;
        case DeviceType.Network:
          if (DialogResult.Cancel == MessageBox.Show(this, "このドライブの解析には時間がかかる可能性があります。", Application.ProductName, MessageBoxButtons.OKCancel, MessageBoxIcon.Information, MessageBoxDefaultButton.Button1))
          {
            return;
          }
          break;
      }

      Visualize(device.RootDirectory);
    }

    private void AbortVisualize()
    {
      if (null != visualizingThread) visualizingThread.Abort();
    }

    private void Visualize(Directory directory)
    {
      if (null == directory) return;

      if (visualizingEvent.WaitOne(0, false))
      {
        MessageBox.Show(this, "他のフォルダを解析中です。", Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Exclamation);

        return;
      }

      // 既に同じディレクトリを解析している場合は、そのVisualizerを再設定する
      Visualizer v = GetVisualizedHistory(directory);

      if (null != v)
      {
        SetVisualizer(v, true);

        return;
      }

      // 開始
      visualizingEvent.Set();

      this.Cursor = Cursors.WaitCursor;

      Visualizer visualizer = new Visualizer(directory);

      SetStatus(0.0f, string.Format("ディレクトリ'{0}'の解析を開始します", directory.Name));

      visualizingThread = new Thread(VisualizeCore);

      visualizingThread.Start(visualizer);

      // タイマーを起動して定期的に監視する
      TimerCallback visualizingState = delegate(object state)
      {
        if (null == visualizer.Status) return;

        uint processed = 0;
        uint total = 0;
        string status = string.Empty;

        lock (visualizer.Status)
        {
          processed = visualizer.Status.Processed;
          total = visualizer.Status.Total;
          status = visualizer.Status.StatusText;
        }

        float progress = 0.0f;

        if (0 == total)
        {
          progress = 0.0f;
        }
        else
        {
          progress = (float)processed / (float)total;
        }

        progress = ((float)visualizer.TaskCompleted + progress) / (float)visualizer.TaskTotal;

        // ステータスを設定
        this.BeginInvoke(new SetStatusDelegate(SetStatus), progress, status);
      };

      visualizingTimer = new System.Threading.Timer(visualizingState, visualizingThread, 0, 250);
    }

    private void VisualizeCore(object parameter)
    {
      string status = string.Empty;
      bool aborted = false;

      Visualizer visualizer = parameter as Visualizer;

      try
      {
        visualizer.AnalysisContext.Parameter.AnalysisDepth = 0;
        visualizer.AnalysisContext.Parameter.PreAnalyze = true;

        viewer.SetGraphingParam(visualizer.GraphingContext.Parameter);

        visualizer.GraphingContext.Parameter.Type = GraphType.Disc;

        viewer.SetRenderingParam(visualizer.RenderingContext.Parameter);

        visualizer.RenderingContext.Parameter.DrawNames = false;
        visualizer.RenderingContext.Parameter.DrawStatistics = false;
        visualizer.RenderingContext.Parameter.InnerDiameterPercent = 0.25f;
        visualizer.RenderingContext.Parameter.OutputWidth = (uint)Math.Min(Screen.PrimaryScreen.Bounds.Width * 4 / 5, Screen.PrimaryScreen.Bounds.Height * 4 / 5);
        visualizer.RenderingContext.Parameter.OutputHeight = visualizer.RenderingContext.Parameter.OutputWidth;

        Stopwatch stopwatch = new Stopwatch();

        stopwatch.Start();

        visualizer.Visualize();

        stopwatch.Stop();

        status = string.Format("完了しました ({0:N2}秒)", stopwatch.Elapsed.TotalSeconds);
      }
      catch (ThreadAbortException)
      {
        // 中断要求
        status = "中断しました";

        aborted = true;

        visualizer = null;
      }
      catch (Exception ex)
      {
        status = "エラーが発生しました: " + ex.Message;

        visualizer = null;

        Console.Error.WriteLine(ex);
      }
      finally
      {
        this.BeginInvoke(new VisualizeFnishedDelegate(VisualizeFinished), visualizer, aborted, status);
      }
    }

    private delegate void VisualizeFnishedDelegate(Visualizer visualized, bool aborted, string status);

    private void VisualizeFinished(Visualizer visualized, bool aborted, string status)
    {
      try
      {
        if (null == visualized)
        {
          if (!aborted)
          {
            MessageBox.Show(this, "解析中にエラーが発生しました。", Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Error);
          }
        }
        else
        {
          SetStatus(1.0f, "結果を表示しています");

          SetVisualizer(visualized, false);
        }

        visualizingTimer.Dispose();

        SetFinalStatus(status);
      }
      finally
      {
        this.Cursor = Cursors.Default;

        visualizingThread = null;

        visualizingEvent.Reset();
      }
    }

    #endregion

    #region "visualization history"

    private List<Visualizer> visualized = new List<Visualizer>();

    private Visualizer GetVisualizedHistory(Directory direcotry)
    {
      foreach (Visualizer v in visualized)
      {
        if (v.AnalysisContext.Parameter.AnalysisDirectory.FullName.Equals(direcotry.FullName))
        {
          return v;
        }
      }

      return null;
    }

    private void AddVisualizedHistory(Visualizer visualizer)
    {
      visualized.Add(visualizer);

      EventHandler clickVisualizedHistoryHandler = delegate(object sender, EventArgs e)
      {
        Visualizer v = (sender as IVisualizerControl).Visualizer;

        SetVisualizer(v, true);
      };

      VisualizedHistoryToolStripButton button = new VisualizedHistoryToolStripButton(visualizer);

      button.Image = Util.Icon.GetIcon(Util.IconType.Folder);
      button.Text = visualizer.AnalysisContext.Parameter.AnalysisDirectory.FullName;
      button.ToolTipText = button.Text;
      button.DisplayStyle = ToolStripItemDisplayStyle.Image;
      button.Click += clickVisualizedHistoryHandler;

      toolStripMain.Items.Add(button);

      VisualizedHistoryToolStripMenuItem item = new VisualizedHistoryToolStripMenuItem(visualizer);

      item.Image = button.Image;
      item.Text = button.Text;
      item.Click += clickVisualizedHistoryHandler;

      toolStripMenuItemVisualize_History.DropDownItems.Add(item);

      // 追加したことにより履歴が5件を超える場合は、一番古い履歴を削除する
      if (visualizationHistoryMax < visualized.Count)
      {
        visualized.RemoveAt(0);

        VisualizedHistoryToolStripButton firstButton = null;

        foreach (ToolStripItem b in toolStripMain.Items)
        {
          if (b is VisualizedHistoryToolStripButton)
          {
            firstButton = b as VisualizedHistoryToolStripButton;

            break;
          }
        }

        if (null != firstButton) toolStripMain.Items.Remove(firstButton);

        VisualizedHistoryToolStripMenuItem firstItem = null;

        foreach (ToolStripItem i in toolStripMenuItemVisualize.DropDownItems)
        {
          if (i is VisualizedHistoryToolStripMenuItem)
          {
            firstItem = i as VisualizedHistoryToolStripMenuItem;

            break;
          }
        }

        if (null != firstItem) toolStripMenuItemVisualize_History.DropDownItems.Remove(firstItem);
      }
    }

    #endregion

    #region "viewing"

    private void SetVisualizer(Visualizer visualizer, bool fromHistory)
    {
      if (visualizer == viewer.Visualizer) return;

      if (fromHistory)
      {
        // TODO: 並べ替え順序が変更されていたら、グラフを作成しなおしたい

        // 履歴として保持されているグラフ画像はすでに破棄されているので再描画する
        viewer.SetRenderingParam(visualizer.RenderingContext.Parameter);

        visualizer.Render(false);
      }

      viewer.SetVisualizer(visualizer);

      // このVisualizerが履歴に存在するかどうか調べる
      if (null != GetVisualizedHistory(visualizer.AnalysisContext.Parameter.AnalysisDirectory)) return;

      // 履歴として記録されていない場合は、履歴として追加する
      AddVisualizedHistory(visualizer);
    }

    private void AscendDirectory()
    {
      if (null != viewer.Graph && null != viewer.Graph.Tree.Parent)
      {
        ViewDirectory(viewer.Graph.Tree.Parent);
      }
    }

    private void ViewDirectory(ITreeGraphElement element)
    {
      ViewDirectory(element.Tree);
    }

    private void ViewDirectory(Tree tree)
    {
      viewer.ViewTree(tree);

      SetStatistics(tree);
    }

    private void GraphRootChanged()
    {
      SetStatistics(viewer.Graph.Tree);

      tabulator.SetElements(viewer.Graph.GetAllElements());

      tabulator.Sort();
    }

    private void SelectedElementChanged()
    {
      viewer.HighlightElements(tabulator.SelectedElements);
    }

    private void SetViewerRenderingOption(bool drawSmallEntries, bool drawHighQuality, Color backgroundColor, Color textColor)
    {
      viewer.DrawHighQuality = drawHighQuality;
      viewer.DrawSmallEntries = drawSmallEntries;
      viewer.SetColor(backgroundColor, textColor);
    }

    private void SetViewerGraphingOption(EntryCompareMode entryCompareMode, bool compareEntryType)
    {
      this.entryCompareMode = entryCompareMode;
      this.compareEntryType = compareEntryType;

      tabulator.CompareEntryType = compareEntryType;

      viewer.EntryCompareMode = entryCompareMode;
      viewer.CompareEntryType = compareEntryType;
    }

    private void SetTabulateMax(int max)
    {
      tabulator.TabulateMax = max;
    }

    #endregion

    #region "viewing history"

    private void ForwardHistory(int step)
    {
      viewer.ForwardHistory(step);

      SetStatistics(viewer.Graph.Tree);
    }

    private void BackHistory(int step)
    {
      viewer.BackHistory(step);

      SetStatistics(viewer.Graph.Tree);
    }

    private void HistoryChanged()
    {
      Tree[] forwardTrees, backwardTrees;
      viewer.GetHistory(out forwardTrees, out backwardTrees);

      toolStripSplitButtonForwardHistory.Enabled = (0 < forwardTrees.Length);
      toolStripSplitButtonBackHistory.Enabled = (0 < backwardTrees.Length);

      // 新しいドロップダウンリストを割り当てる
      toolStripSplitButtonForwardHistory.DropDownItems.Clear();

      foreach (Tree tree in forwardTrees)
      {
        toolStripSplitButtonForwardHistory.DropDownItems.Add(tree.RootDirectory.FullName, null, delegate(object sender, EventArgs e)
        {
          ForwardHistory(toolStripSplitButtonForwardHistory.DropDownItems.IndexOf(sender as ToolStripItem) + 1);
        });
      }

      toolStripSplitButtonBackHistory.DropDownItems.Clear();

      foreach (Tree tree in backwardTrees)
      {
        toolStripSplitButtonBackHistory.DropDownItems.Add(tree.RootDirectory.FullName, null, delegate(object sender, EventArgs e)
        {
          BackHistory(toolStripSplitButtonBackHistory.DropDownItems.IndexOf(sender as ToolStripItem) + 1);
        });
      }
    }

    #endregion

    #region "searching"

    private void SearchByWildcard(string wildcard)
    {
      // ワイルドカードを正規表現に置き換える
      wildcard = wildcard.Replace(@"\", @"\\");
      wildcard = wildcard.Replace(@".", @"\.");
      wildcard = wildcard.Replace("*", ".*");
      wildcard = wildcard.Replace("?", ".");

      SearchByRegex("^" + wildcard + "$");
    }

    private void SearchByRegex(string regexString)
    {
      if (null == viewer.Graph) return;

      Regex regex = null;

      try
      {
        regex = new Regex(regexString, RegexOptions.Singleline);
      }
      catch (ArgumentException ex)
      {
        MessageBox.Show(this, "正規表現に誤りがあります。" + Environment.NewLine + ex.Message, Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Exclamation);

        return;
      }

      IGraphElement[] matched = new IGraphElement[] { };
      
      try
      {
        this.Cursor = Cursors.WaitCursor;

        matched = viewer.Graph.SearchAllElement(regex, delegate(Regex regexp, IGraphElement element)
        {
          return regexp.IsMatch(element.Entry.Name);
        });

        viewer.HighlightElements(matched);

        tabulator.SetElements(matched);

        tabulator.Sort();
      }
      finally
      {
        this.Cursor = Cursors.Default;
      }

      if (0 == matched.Length)
      {
        SetFinalStatus("該当するファイル・フォルダはありません");
      }
      else
      {
        SetFinalStatus(string.Format("該当するファイル・フォルダが{0}個見つかりました", matched.Length));
      }
    }

    #endregion

    #region "saving"

    private void SaveResult(FormatType type)
    {
      if (null == viewer.Visualizer) return;

      // 拡張子と表記
      string[,] extensions;

      switch(type)
      {
        case FormatType.Graph:
          extensions = new string[,]
          {
            {"png", "透過PNG形式"},
            {"jpg", "JPEG形式"},
            {"bmp", "ビットマップ形式(32bit)"},
          };
          break;

        case FormatType.Xml:
          extensions = new string[,]
          {
            {"xml", "XMLファイル"},
          };
          break;

        case FormatType.Csv:
          extensions = new string[,]
          {
            {"csv", "CSVファイル"},
          };
          break;

        default:
          return;
      }

      // フィルタ文字列を作成
      string filter = string.Empty;

      for (int index = 0; ; index++)
      {
        filter += string.Format("{1} (*.{0})|*.{0}", extensions[index, 0], extensions[index, 1]);

        if (extensions.GetLength(0) - 1 <= index) break;

        filter += "|";
      }

      // デフォルトのファイル名
      string filename = viewer.Graph.Tree.RootDirectory.Name;

      foreach (char c in System.IO.Path.GetInvalidFileNameChars())
      {
        filename = filename.Replace(c, '_');
      }

      SaveFileDialog dialog = new SaveFileDialog();

      dialog.AddExtension = true;
      dialog.DefaultExt = extensions[0, 0];
      dialog.FileName = string.Format("{0}.{1}", filename, extensions[0, 0]);
      dialog.Filter = filter;
      dialog.FilterIndex = 0;
      dialog.InitialDirectory = Environment.GetFolderPath(Environment.SpecialFolder.Desktop);
      dialog.OverwritePrompt = true;
      dialog.RestoreDirectory = false;
      dialog.SupportMultiDottedExtensions = true;
      dialog.ValidateNames = true;

      // ダイアログを表示
      if (dialog.ShowDialog(this) != DialogResult.OK) return;

      // 別スレッドで保存処理を開始
      Thread t = new Thread(SaveResultThread);

      if (type == FormatType.Graph)
      {
        t.Start(new object[] { type, viewer.Visualizer, dialog.FileName });
      }
      else
      {
        t.Start(new object[] { type, viewer.Visualizer, dialog.FileName, tabulator.TabulatingElements });
      }

      this.Cursor = Cursors.WaitCursor;
    }

    private void SaveResultThread(object param)
    {
      object[] parameter = param as object[];

      FormatType type = (FormatType)parameter[0];
      Visualizer visualizer = parameter[1] as Visualizer;
      string filename = parameter[2] as string;

      string status = string.Empty;
      FileStream stream = null;
      bool succeeded = false;

      try
      {
        stream = new FileStream(filename, FileMode.Create, FileAccess.Write, FileShare.None);

        if (type == FormatType.Graph)
        {
          SaveResultAsImage(visualizer, stream);
        }
        else
        {
          SaveResultAsText(visualizer, stream, type, parameter[3] as IGraphElement[]);
        }

        status = string.Format("保存しました({0})", stream.Name);

        succeeded = true;
      }
      catch (ThreadAbortException)
      {
        // 中断要求
        status = "中断しました";
      }
      catch (Exception ex)
      {
        status = "エラーが発生しました: " + ex.Message;

        Console.Error.WriteLine(ex);
      }
      finally
      {
        if (stream != null)
        {
          stream.Close();

          stream = null;
        }

        this.BeginInvoke(new SaveResultFinishedDelegate(SaveResultFinished), status, succeeded);
      }
    }

    private void SaveResultAsImage(Visualizer visualizer, FileStream stream)
    {
      // Rendererで描画する
      RenderingContext context = new RenderingContext();

      context.Parameter.DrawNames = true;
      context.Parameter.DrawSmallEntries = true;
      context.Parameter.DrawStatistics = true;
      context.Parameter.BackgroundColor = visualizer.RenderingContext.Parameter.BackgroundColor;
      context.Parameter.TextColor = visualizer.RenderingContext.Parameter.TextColor;
      context.Parameter.Graph = visualizer.GraphingContext.Result.Graph;
      context.Parameter.InnerDiameterPercent = 0.25f;
      context.Parameter.OutputWidth = 1280;
      context.Parameter.OutputHeight = 1280;

      (new Renderer(context)).Render();

      // ストリームに書き込む
      ImageFormat format;

      switch (System.IO.Path.GetExtension(stream.Name).ToLower())
      {
        case ".png":
          format = ImageFormat.Png;
          break;

        case ".jpg":
          format = ImageFormat.Jpeg;
          break;

        case ".bmp":
          format = ImageFormat.Bmp;
          break;

        default:
          // もし上記以外の拡張子で指定されたら
          format = ImageFormat.Png;
          break;
      }

      context.Result.GraphImage.Save(stream, format);
    }

    private void SaveResultAsText(Visualizer visualizer, FileStream stream, FormatType type, IGraphElement[] formattingElements)
    {
      // Formatterで所定の形式に整形する
      visualizer.FormattingContext.Parameter.Graph = visualizer.GraphingContext.Result.Graph;
      visualizer.FormattingContext.Parameter.FormattingElements = formattingElements;

      switch (type)
      {
        case FormatType.Xml:
          visualizer.FormattingContext.Parameter.Writer = new StreamWriter(stream, Encoding.UTF8);
          visualizer.FormattingContext.Parameter.Type = Format.FormatType.Xml;
          break;

        case FormatType.Csv:
          visualizer.FormattingContext.Parameter.Writer = new StreamWriter(stream, Encoding.Default);
          visualizer.FormattingContext.Parameter.Type = Format.FormatType.Csv;
          break;
      }

      visualizer.Format(false);
    }

    private delegate void SaveResultFinishedDelegate(string status, bool succeeded);

    private void SaveResultFinished(string status, bool succeeded)
    {
      try
      {
        if (!succeeded)
        {
          MessageBox.Show(this, "保存できませんでした。", Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Error);
        }

        SetFinalStatus(status);
      }
      finally
      {
        this.Cursor = Cursors.Default;

        // TODO: 書き込み中に失敗した場合に残るファイルは気にしないことにする
      }
    }

    #endregion

    #region "status/statistics"

    private delegate void SetStatusDelegate(float progress, string status);

    private void SetStatus(float progress, string statusText)
    {
      if (progress < 0.0f) progress = 0.0f;
      if (1.0f < progress) progress = 1.0f;

      if (!toolStripProgressBarStatus.Visible)
      {
        toolStripProgressBarStatus.Visible = true;
        toolStripProgressBarStatus.Minimum = 0;
        toolStripProgressBarStatus.Maximum = 100;
      }

      toolStripProgressBarStatus.Value = (int)(100.0f * progress);

      toolStripStatusLabelStatus.Text = statusText;
    }

    private void SetFinalStatus(string statusText)
    {
      if (toolStripProgressBarStatus.Visible)
      {
        toolStripProgressBarStatus.Visible = false;
      }

      toolStripStatusLabelStatus.Text = statusText;
    }

    private void SetStatistics(Tree tree)
    {
      textBoxStatistics.Text = string.Format("{0}{1}全エントリ数: {2:N0} (ファイル:{3:N0} フォルダ:{4:N0}) 総容量: {5}({6})",
        tree.RootDirectory.FullName,
        Environment.NewLine,
        tree.GetAllEntryCount(),
        tree.GetAllFileCount(),
        tree.GetAllSubTreeCount(),
        tree.RootDirectory.Size.ToBinaryPrefixedString(),
        tree.RootDirectory.Size.ToNumericString());
    }

    private void SetStatistics(IGraphElement element)
    {
      if (element is ITreeGraphElement)
      {
        SetStatistics((element as ITreeGraphElement).Tree);
      }
      else
      {
        File file = element.Entry as File;
        FileSystem.Size size = file.Size;

        textBoxStatistics.Text = string.Format("{0}{1}{2}({3})",
          file.FullName,
          Environment.NewLine,
          size.ToBinaryPrefixedString(),
          size.ToNumericString());
      }
    }

    #endregion
  }
}