using System;
using System.Diagnostics;
using System.Collections.Generic;
using System.Windows.Forms;

namespace Smdn.Applications.DirectoryAnalyzer.UI
{
  using FileSystem;
  using Graphing;

  public class TabulatorMouseEventArgs : MouseEventArgs
  {
    private IGraphElement element = null;

    public IGraphElement Element
    {
      get
      {
        return this.element;
      }
    }

    public TabulatorMouseEventArgs(IGraphElement element, MouseEventArgs baseEvent)
      : base(baseEvent.Button, baseEvent.Clicks, baseEvent.X, baseEvent.Y, baseEvent.Delta)
    {
      this.element = element;
    }
  }

  public delegate void TabulatorMouseEventHandler(object sender, TabulatorMouseEventArgs e);

  public partial class Tabulator
  {
    public enum EntryCompareOrder : int
    {
      Ascending,
      Descending,
    }

    private ListView tabulatingListView = null;
    private IGraphElement[] tabulatingElements = null;
    private int tabulateMax = 30;
    private EntryCompareMode entryCompareMode = EntryCompareMode.BySize;
    private EntryCompareOrder entryCompareOrder = EntryCompareOrder.Ascending;
    private bool compareEntryType = true;
    private List<ListViewItem> lastSelectedItems = new List<ListViewItem>();

    public event TabulatorMouseEventHandler ItemClicked = null;
    public event EventHandler SelectedElementChanged = null;

    public ListView TabulatingListView
    {
      get
      {
        return this.tabulatingListView;
      }
    }

    public int TabulateMax
    {
      get
      {
        return this.tabulateMax;
      }
      set
      {
        if (value < 0) throw new ArgumentOutOfRangeException("TabulateMax");

        if (value != this.tabulateMax)
        {
          this.tabulateMax = value;

          if (null != tabulatingElements)
          {
            UpdateListItem();
          }
        }
      }
    }

    public IGraphElement[] TabulatingElements
    {
      get
      {
        return this.tabulatingElements;
      }
    }

    public IGraphElement[] SelectedElements
    {
      get
      {
        List<IGraphElement> selected = new List<IGraphElement>();

        foreach(ListViewItem item in this.tabulatingListView.SelectedItems)
        {
          if (item is ElementListViewItem)
          {
            selected.Add((item as ElementListViewItem).Element);
          }
        }

        return selected.ToArray();
      }
    }

    public bool CompareEntryType
    {
      get
      {
        return this.compareEntryType;
      }
      set
      {
        if (value != this.compareEntryType)
        {
          this.compareEntryType = value;

          Sort();
        }
      }
    }

    public Tabulator(ListView tabulatingListView)
    {
      this.tabulatingListView = tabulatingListView;

      InitializeListView();

      InitializeHandler();
    }

    private void InitializeListView()
    {
      tabulatingListView.Activation = ItemActivation.Standard;
      tabulatingListView.AllowColumnReorder = true;
      tabulatingListView.CheckBoxes = false;
      tabulatingListView.HeaderStyle = ColumnHeaderStyle.Clickable;
      tabulatingListView.View = View.Details;
      tabulatingListView.FullRowSelect = true;
      tabulatingListView.MultiSelect = true;
      tabulatingListView.Scrollable = true;
      tabulatingListView.LabelEdit = false;
      tabulatingListView.ListViewItemSorter = null;
      tabulatingListView.ShowItemToolTips = true;

      tabulatingListView.Columns.Clear();
      tabulatingListView.Columns.Add(new EntryListViewColumn("名前", 120, EntryCompareMode.ByName));
      tabulatingListView.Columns.Add("種類", 24);
      tabulatingListView.Columns.Add(new EntryListViewColumn("容量[%]", 64, EntryCompareMode.BySize));
      tabulatingListView.Columns.Add(new EntryListViewColumn("容量[Bytes]", 64, EntryCompareMode.BySize));
      tabulatingListView.Columns.Add(new EntryListViewColumn("フルパス", 540, EntryCompareMode.ByFullName));
    }

    private void InitializeHandler()
    {
      tabulatingListView.MouseUp += delegate(object sender, MouseEventArgs e)
      {
        ElementListViewItem item = tabulatingListView.GetItemAt(e.X, e.Y) as ElementListViewItem;

        if (null != item)
        {
          OnItemClicked(new TabulatorMouseEventArgs(item.Element, e));
        }

        // 選択されているアイテムが変わった場合
        if (IsSelectedItemChanged())
        {
          // SelectedItemChangedイベントは、アイテムの内容が変わる度に発生するので使わない
          // (複数選択→1件選択だと、減った選択数分SelectedItemChangedイベントが発生する)
          // キー入力での選択に対応できないことについては、割り切る
          OnSelectedElementChanged(EventArgs.Empty);
        }
      };
        
      tabulatingListView.ColumnClick += delegate(object sender, ColumnClickEventArgs e)
      {
        EntryListViewColumn column = (sender as ListView).Columns[e.Column] as EntryListViewColumn;

        if (null != column)
        {
          if (column.CompareMode == this.entryCompareMode)
          {
            // ソート方法が同じ場合は、昇降順を逆にする
            if (this.entryCompareOrder == EntryCompareOrder.Ascending)
            {
              Sort(this.entryCompareMode, EntryCompareOrder.Descending);
            }
            else
            {
              Sort(this.entryCompareMode, EntryCompareOrder.Ascending);
            }
          }
          else
          {
            // ソート方法が異なる場合は、新しいソート方法で並べ替える
            Sort(column.CompareMode, EntryCompareOrder.Ascending);
          }
        }
      };
    }

    public void SetElements(IGraphElement[] elements)
    {
      tabulatingElements = elements;
    }

    public void Sort()
    {
      Sort(entryCompareMode, entryCompareOrder);
    }

    public void Sort(EntryCompareMode sortMode, EntryCompareOrder sortOrder)
    {
      if (null == tabulatingElements) return;

      // 指定の方法で並び替える
      Comparison<Entry> comparer = Comparison.GetComparer(sortMode, compareEntryType);

      if (null != comparer)
      {
        Array.Sort<IGraphElement>(tabulatingElements, delegate(IGraphElement x, IGraphElement y)
        {
          return comparer(x.Entry, y.Entry);
        });
      }

      if (sortOrder == EntryCompareOrder.Descending)
      {
        Array.Reverse(tabulatingElements);
      }

      UpdateListItem();

      entryCompareMode = sortMode;
      entryCompareOrder = sortOrder;
    }

    private void UpdateListItem()
    {
      IGraphElement[] viewingElements;
      int cutElementCount = 0;

      Stopwatch stopwatch = new Stopwatch();

      Console.WriteLine("Updating tabulation items...");

      stopwatch.Start();

      // tabulateMaxで指定されている件数に絞る
      if (0 != tabulateMax && tabulateMax < tabulatingElements.Length)
      {
        viewingElements = Array.CreateInstance(typeof(IGraphElement), tabulateMax) as IGraphElement[];

        Array.Copy(tabulatingElements, viewingElements, tabulateMax);

        cutElementCount = tabulatingElements.Length - tabulateMax;
      }
      else
      {
        viewingElements = tabulatingElements;
      }

      // ListViewの項目を作成する
      try
      {
        tabulatingListView.BeginUpdate();

        tabulatingListView.Items.Clear();

        foreach (IGraphElement element in viewingElements)
        {
          if (element.Entry is Directory)
          {
            tabulatingListView.Items.Add(new DirectoryListViewItem(element));
          }
          else
          {
            tabulatingListView.Items.Add(new FileListViewItem(element));
          }
        }

        if (0 != cutElementCount)
        {
          tabulatingListView.Items.Add(new ListViewItem(string.Format("他に{0}個のファイル・フォルダがあります", cutElementCount)));
        }
      }
      finally
      {
        tabulatingListView.EndUpdate();
      }

      stopwatch.Stop();

      Console.WriteLine("  Elaplsed: {0}", stopwatch.Elapsed);
      Console.WriteLine();
    }

    protected virtual void OnItemClicked(TabulatorMouseEventArgs e)
    {
      if (null != ItemClicked) ItemClicked(this, e);
    }

    protected virtual void OnSelectedElementChanged(EventArgs e)
    {
      if (null != SelectedElementChanged) SelectedElementChanged(this, e);
    }

    private bool IsSelectedItemChanged()
    {
      List<ListViewItem> selectedItems = new List<ListViewItem>();

      bool changed = false;

      foreach (ListViewItem item in tabulatingListView.SelectedItems)
      {
        if (!lastSelectedItems.Contains(item)) changed = true;

        selectedItems.Add(item);
      }

      return changed;
    }
  }
}
