using System;
using System.Collections.Generic;
using System.Drawing;

namespace Smdn.Applications.DirectoryAnalyzer.UI
{
  using Analysis;
  using FileSystem;
  using Graphing;
  using Render;

  public partial class Viewer
  {
    private class ViewingHistoryEntry
    {
      public readonly Tree Tree = null;
      public Graph Graph = null;
      public Bitmap GraphImage = null;

      public ViewingHistoryEntry(Tree tree, Graph graph, Bitmap graphImage)
      {
        this.Tree = tree;
        this.Graph = graph;
        this.GraphImage = graphImage;
      }
    }

    private class ViewingHistory
    {
      private int historyMax = 0;
      private int currentEntry = -1;
      private List<ViewingHistoryEntry> entries = null;

      public ViewingHistoryEntry Current
      {
        get
        {
          if (currentEntry < 0)
          {
            return null;
          }
          else
          {
            if (entries.Count == 0)
            {
              return null;
            }
            else
            {
              return entries[currentEntry];
            }
          }
        }
      }

      public ViewingHistory(int historyMax)
      {
        this.historyMax = historyMax;

        this.entries = new List<ViewingHistoryEntry>(historyMax);

        Clear();
      }

      public void Add(ViewingHistoryEntry entry)
      {
        if (entries.Count == historyMax)
        {
          // ヒストリの最大数を超える場合は、一番古いヒストリを削除する
          ViewingHistoryEntry oldestEntry = entries[entries.Count - 1];

          entries.Remove(oldestEntry);

          oldestEntry.GraphImage.Dispose();
        }

        // Backしている状態でAddするときは、Forward分を削除する
        // (以後Forwardできないようにする)
        for (; 0 < currentEntry; currentEntry--)
        {
          entries.RemoveAt(0);
        }

        currentEntry = 0;

        entries.Insert(currentEntry, entry);
      }

      public void Clear()
      {
        Clear(false);
      }

      public void Clear(bool exceptCurrent)
      {
        ViewingHistoryEntry current = null;

        if (exceptCurrent)
        {
          current = this.Current;

          entries.Remove(current);
        }

        foreach (ViewingHistoryEntry entry in entries)
        {
          entry.GraphImage.Dispose();
        }

        entries.Clear();

        currentEntry = -1;

        if (current != null)
        {
          Add(current);
        }
      }

      public void Forward()
      {
        if (0 == entries.Count) currentEntry = -1;

        if (0 < currentEntry)
        {
          currentEntry -= 1;
        }
      }

      public void Back()
      {
        if (0 == entries.Count) currentEntry = -1;

        if (currentEntry < entries.Count - 1)
        {
          currentEntry += 1;
        }
      }

      public void GetHistory(out ViewingHistoryEntry[] forwards, out ViewingHistoryEntry[] backwards)
      {
        if (0 == entries.Count)
        {
          forwards = new ViewingHistoryEntry[] { };
          backwards = new ViewingHistoryEntry[] { };
        }

        ViewingHistoryEntry[] e = entries.ToArray();

        forwards = Array.CreateInstance(typeof(ViewingHistoryEntry), currentEntry) as ViewingHistoryEntry[];

        Array.Copy(e, forwards, forwards.Length);

        Array.Reverse(forwards);

        if (entries.Count == 1)
        {
          backwards = new ViewingHistoryEntry[] { };
        }
        else
        {
          backwards = Array.CreateInstance(typeof(ViewingHistoryEntry), entries.Count - (currentEntry + 1)) as ViewingHistoryEntry[];
        }

        Array.Copy(e, currentEntry + 1, backwards, 0, backwards.Length);
      }

      public void GetHistory(out Tree[] forwards, out Tree[] backwards)
      {
        ViewingHistoryEntry[] forwardEntries, backwardEntries;

        GetHistory(out forwardEntries, out backwardEntries);

        forwards = Array.CreateInstance(typeof(Tree), forwardEntries.Length) as Tree[];

        for (int index = 0; index < forwardEntries.Length; index++)
        {
          forwards[index] = forwardEntries[index].Tree;
        }

        backwards = Array.CreateInstance(typeof(Tree), backwardEntries.Length) as Tree[];

        for (int index = 0; index < backwardEntries.Length; index++)
        {
          backwards[index] = backwardEntries[index].Tree;
        }
      }
    }
  }
}