using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.Drawing.Text;
using System.Text.RegularExpressions;
using System.Windows.Forms;

namespace Smdn.Applications.DirectoryAnalyzer.UI
{
  using Analysis;
  using FileSystem;
  using Graphing;
  using Render;

  public class ViewerMouseEventArgs : MouseEventArgs
  {
    private IGraphElement element = null;

    public IGraphElement Element
    {
      get
      {
        return this.element;
      }
    }

    public ViewerMouseEventArgs(IGraphElement element, MouseEventArgs baseEvent)
      : base(baseEvent.Button, baseEvent.Clicks, baseEvent.X, baseEvent.Y, baseEvent.Delta)
    {
      this.element = element;
    }
  }

  public delegate void ViewerMouseEventHandler(object sender, ViewerMouseEventArgs e);

  public class ViewerStatusEventArgs : EventArgs
  {
    private string status = string.Empty;

    public string Status
    {
      get
      {
        return this.status;
      }
    }

    public ViewerStatusEventArgs(string status)
    {
      this.status = status;
    }
  }

  public delegate void ViewerStatusEventHandler(object sender, ViewerStatusEventArgs e);

  public partial class Viewer
  {
    private const int viewingHistoryMax = 5;

    private Visualizer visualizer = null;
    private Control viewingControl = null;
    private IGraphElement elementAtCursor = null;
    private Renderer elementRetrievingRenderer = null;
    private PartialGraphImage highlightedPartialGraph = null;
    private float viewingGraphRatio = 0.0f;
    private RectangleF viewingGraphBounds = Rectangle.Empty;
    private ViewingHistory viewingHistory = new ViewingHistory(viewingHistoryMax);
    private bool drawSmallEntries = false;
    private bool drawHighQuality = true;
    private Color backgroundColor = Color.White;
    private Color textColor = Color.Black;
    private EntryCompareMode entryCompareMode = EntryCompareMode.BySize;
    private bool compareEntryType = true;
    private Comparison<Entry> entryComparer = null;

    public event ViewerMouseEventHandler ElementAtCursorChanged = null;
    public event ViewerMouseEventHandler ElementClicked = null;
    public event ViewerStatusEventHandler StatusChanged = null;
    public event EventHandler GraphRootChanged = null;
    public event EventHandler HistoryChanged = null;

    public Visualizer Visualizer
    {
      get
      {
        return this.visualizer;
      }
    }

    public Control ViewingControl
    {
      get
      {
        return this.viewingControl;
      }
    }

    public Graph Graph
    {
      get
      {
        if (null == this.viewingHistory.Current)
        {
          return null;
        }
        else
        {
          return this.viewingHistory.Current.Graph;
        }
      }
    }

    public bool DrawSmallEntries
    {
      get
      {
        return this.drawSmallEntries;
      }
      set
      {
        if (this.drawSmallEntries != value)
        {
          this.drawSmallEntries = value;

          RenderingParameterChanged();
        }
      }
    }

    public bool DrawHighQuality
    {
      get
      {
        return this.drawHighQuality;
      }
      set
      {
        if (this.drawHighQuality != value)
        {
          this.drawHighQuality = value;

          RenderingParameterChanged();
        }
      }
    }

    public Color BackgroundColor
    {
      get
      {
        return this.backgroundColor;
      }
    }

    public Color TextColor
    {
      get
      {
        return this.textColor;
      }
    }

    public void SetColor(Color background, Color text)
    {
      if (this.backgroundColor != background || this.textColor != text)
      {
        this.backgroundColor = background;
        this.textColor = text;

        RenderingParameterChanged();
      }
    }

    public EntryCompareMode EntryCompareMode
    {
      get
      {
        return this.entryCompareMode;
      }
      set
      {
        if ( value != this.entryCompareMode)
        {
          this.entryCompareMode = value;

          GraphingParameterChanged();
        }
      }
    }

    public bool CompareEntryType
    {
      get
      {
        return this.compareEntryType;
      }
      set
      {
        if (value != this.compareEntryType)
        {
          this.compareEntryType = value;

          GraphingParameterChanged();
        }
      }
    }

    /// <summary>
    /// コンストラクタ
    /// </summary>
    public Viewer(Control viewingControl)
    {
      this.viewingControl = viewingControl;

      this.entryComparer = Comparison.GetComparer(this.entryCompareMode, compareEntryType);

      InitializeHandler();
    }

    private void InitializeHandler()
    {
      // リサイズ時に描画する
      viewingControl.Resize += delegate(object sender, EventArgs e)
      {
        MeasureViewingSize();

        viewingControl.Invalidate();
      };

      // 描画処理
      viewingControl.Paint += delegate(object sender, PaintEventArgs e)
      {
        RenderView(e.Graphics);
      };

      // マウスの移動
      viewingControl.MouseMove += delegate(object sender, MouseEventArgs e)
      {
        Point point = new Point(e.X, e.Y);

        IGraphElement element = GetElementAt(point);

        if (this.elementAtCursor != element)
        {
          this.elementAtCursor = element;

          if (null == element)
          {
            OnEntryAtCursorChanged(new ViewerMouseEventArgs(null, e));
          }
          else
          {
            OnEntryAtCursorChanged(new ViewerMouseEventArgs(element, e));
          }
        }
      };

      // 左右クリック
      viewingControl.MouseClick += delegate(object sender, MouseEventArgs e)
      {
        if (null != this.elementAtCursor)
        {
          OnEntryClicked(new ViewerMouseEventArgs(this.elementAtCursor, e));
        }
      };
    }

    /// <summary>
    /// ツリーを見る
    /// </summary>
    /// <param name="tree">見る対象のツリー</param>
    /// <param name="updateGraph">ツリーを表示する前にグラフを再作成するかどうか</param>
    /// <param name="updateRendered">ツリーを表示する前にグラフを再描画するかどうか</param>
    /// <param name="addToHistory">作成・描画したグラフをヒストリに追加するかどうか</param>
    /// <param name="updateCurrentHistory">ヒストリに追加しない場合、現在のヒストリの状態を更新するかどうか</param>
    private void ViewTree(Tree tree, bool updateGraph, bool updateRendered, bool addToHistory, bool updateCurrentHistory)
    {
      Stopwatch stopwatch = new Stopwatch();

      stopwatch.Start();

      Console.WriteLine("Selected tree '{0}'", tree.RootDirectory.FullName);

      // ハイライトされた部分は無効となるため破棄する
      if (null != highlightedPartialGraph)
      {
        lock (highlightedPartialGraph)
        {
          highlightedPartialGraph.Dispose();

          highlightedPartialGraph = null;
        }
      }

      // グラフの再作成・再描画
      if (updateGraph || updateRendered)
      {
        OnStatusChanged(new ViewerStatusEventArgs("グラフを描画しています..."));

        // グラフを再作成する
        if (updateGraph)
        {
          SetGraphingParam(visualizer.GraphingContext.Parameter);

          visualizer.GraphingContext.Parameter.Tree = tree;

          visualizer.Graph(false);

          // 位置取得用のRendererは無効となるため再作成させる
          elementRetrievingRenderer = null;
        }

        // グラフを再描画する
        if (updateRendered)
        {
          SetRenderingParam(visualizer.RenderingContext.Parameter);

          visualizer.RenderingContext.Parameter.Graph = visualizer.GraphingContext.Result.Graph;

          visualizer.Render(false);
        }
      }

      // 位置取得用のRendererを再作成
      if (null == elementRetrievingRenderer)
      {
        elementRetrievingRenderer = new Renderer(visualizer.RenderingContext);
      }

      if (addToHistory)
      {
        // 新しいエントリをヒストリに追加
        viewingHistory.Add(new ViewingHistoryEntry(
          tree,
          visualizer.GraphingContext.Result.Graph,
          visualizer.RenderingContext.Result.GraphImage));

        OnHistoryChanged(EventArgs.Empty);
      }
      else if (updateCurrentHistory)
      {
        // 現在のヒストリの状態を更新
        lock(viewingHistory.Current)
        {
          viewingHistory.Current.Graph = visualizer.GraphingContext.Result.Graph;
          viewingHistory.Current.GraphImage.Dispose();
          viewingHistory.Current.GraphImage = visualizer.RenderingContext.Result.GraphImage;
        }
      }

      // グラフのルートの変更を通知
      OnGraphRootChanged(EventArgs.Empty);

      // 再描画
      MeasureViewingSize();

      viewingControl.Invalidate();

      // ステータスを更新
      stopwatch.Stop();

      OnStatusChanged(new ViewerStatusEventArgs(
        string.Format("描画が完了しました ({0:N2}秒, {1})",
          stopwatch.Elapsed.TotalSeconds,
          tree.RootDirectory.FullName)
        ));
    }

    /// <summary>
    /// ヒストリに設定されているツリーを見る
    /// </summary>
    /// <param name="entry">ヒストリのエントリ</param>
    /// <param name="updateGraph">グラフを再作成するかどうか</param>
    /// <param name="updateRendered">グラフを再描画するかどうか</param>
    private void ViewHistory(ViewingHistoryEntry entry, bool updateGraph, bool updateRendered)
    {
      // 各タスクのパラメータを復元
      visualizer.GraphingContext.Parameter.Tree = viewingHistory.Current.Tree;
      visualizer.RenderingContext.Parameter.Graph = viewingHistory.Current.Graph;

      if (updateGraph || updateRendered)
      {
        ViewTree(entry.Tree, updateGraph, updateRendered, false, true);
      }
      else
      {
        // 状態を更新しない場合は各タスクの結果も復元
        visualizer.GraphingContext.Result.Graph = viewingHistory.Current.Graph;
        visualizer.RenderingContext.Result.GraphImage = viewingHistory.Current.GraphImage;

        ViewTree(entry.Tree, false, false, false, false);
      }

      OnHistoryChanged(EventArgs.Empty);
    }

    /// <summary>
    /// グラフの描画領域を再計算する
    /// </summary>
    private void MeasureViewingSize()
    {
      if (null == viewingHistory.Current) return;

      // グラフ全体を描画できる拡大縮小率と、描画領域を求める
      Bitmap graphImage = viewingHistory.Current.GraphImage;

      float tanGraphImage = graphImage.Height / (float)graphImage.Width;
      float tanPictureBox = viewingControl.Height / (float)viewingControl.Width;

      if (tanGraphImage < tanPictureBox)
      {
        viewingGraphRatio = viewingControl.Width / (float)graphImage.Width;

        viewingGraphBounds = new RectangleF(0.0f, 0.0f, viewingControl.Width, graphImage.Height * viewingGraphRatio);
        viewingGraphBounds.Y = (viewingControl.Height - graphImage.Height * viewingGraphRatio) * 0.5f;
      }
      else
      {
        viewingGraphRatio = viewingControl.Height / (float)graphImage.Height;

        viewingGraphBounds = new RectangleF(0.0f, 0.0f, graphImage.Width * viewingGraphRatio, viewingControl.Height);
        viewingGraphBounds.X = (viewingControl.Width - graphImage.Width * viewingGraphRatio) * 0.5f;
      }
    }

    /// <summary>
    /// グラフの描画を行う
    /// </summary>
    private void RenderView(Graphics g)
    {
      g.Clear(SystemColors.Control);

      if (null == viewingHistory.Current) return;

      // 描画品質を設定する
      if (drawHighQuality)
      {
        g.InterpolationMode = InterpolationMode.HighQualityBicubic;
      }
      else
      {
        g.InterpolationMode = InterpolationMode.NearestNeighbor;
      }

      // 描画する
      switch (viewingHistory.Current.Graph.Type)
      {
        case GraphType.Disc:
          g.DrawImage(viewingHistory.Current.GraphImage, viewingGraphBounds);
          break;

        default:
          break;
      }

      // ハイライトされた部分を上書きする
      if (null != highlightedPartialGraph)
      {
        g.DrawImage(
          highlightedPartialGraph.Image,
          viewingGraphBounds.X + highlightedPartialGraph.ImageOffsetX * viewingGraphRatio,
          viewingGraphBounds.Y + highlightedPartialGraph.ImageOffsetY * viewingGraphRatio,
          highlightedPartialGraph.Image.Width * viewingGraphRatio,
          highlightedPartialGraph.Image.Height * viewingGraphRatio);
      }
    }

    /// <summary>
    /// 現在の設定に基づいてグラフ描画に関するパラメータを設定する
    /// </summary>
    public void SetRenderingParam(RenderingContext.RenderingParameter param)
    {
      param.DrawSmallEntries = drawSmallEntries;

      if (drawHighQuality)
      {
        param.SmoothingMode = SmoothingMode.AntiAlias;
        param.TextRenderingHint = TextRenderingHint.AntiAlias;
        param.InterpolationMode = InterpolationMode.HighQualityBicubic;
      }
      else
      {
        param.SmoothingMode = SmoothingMode.HighSpeed;
        param.TextRenderingHint = TextRenderingHint.SystemDefault;
        param.InterpolationMode = InterpolationMode.NearestNeighbor;
      }

      param.BackgroundColor = backgroundColor;
      param.TextColor = textColor;
    }

    /// <summary>
    /// 現在の設定に基づいてグラフ作成に関するパラメータを設定する
    /// </summary>
    public void SetGraphingParam(GraphingContext.GraphingParameter param)
    {
      param.Comparer = entryComparer;
    }

    /// <summary>
    /// 新しいVisualizerを設定する
    /// </summary>
    public void SetVisualizer(Visualizer visualizer)
    {
      if (visualizer == this.visualizer) return;

      this.visualizer = visualizer;

      // 位置取得用のRendererは無効となるため再作成させる
      elementRetrievingRenderer = null;

      // ヒストリをクリアする
      viewingHistory.Clear();

      ViewTree(visualizer.GraphingContext.Parameter.Tree, false, false, true, false);
    }

    /// <summary>
    /// ヒストリを進む
    /// </summary>
    public void ForwardHistory(int step)
    {
      if (null == visualizer) return;
      if (step <= 0) throw new InvalidOperationException("step");

      for (int s = 0; s < step; s++)
      {
        viewingHistory.Forward();
      }

      if (null != viewingHistory.Current)
      {
        ViewHistory(viewingHistory.Current, false, false);
      }
    }

    /// <summary>
    /// ヒストリを遡る
    /// </summary>
    public void BackHistory(int step)
    {
      if (null == visualizer) return;
      if (step <= 0) throw new InvalidOperationException("step");

      for (int s = 0; s < step; s++)
      {
        viewingHistory.Back();
      }

      if (null != viewingHistory.Current)
      {
        ViewHistory(viewingHistory.Current, false, false);
      }
    }

    /// <summary>
    /// ヒストリを取得する
    /// </summary>
    public void GetHistory(out Tree[] forwards, out Tree[] backwards)
    {
      if (null == visualizer)
      {
        forwards = new Tree[] { };
        backwards = new Tree[] { };
        return;
      }

      viewingHistory.GetHistory(out forwards, out backwards);
    }

    /// <summary>
    /// 指定された要素をハイライトする
    /// </summary>
    public void HighlightElements(IGraphElement[] elements)
    {
      if (null == elements) return;

      // ハイライトされた部分は無効となるため破棄する
      if (null != highlightedPartialGraph)
      {
        lock (highlightedPartialGraph)
        {
          highlightedPartialGraph.Dispose();

          highlightedPartialGraph = null;
        }
      }

      if (0 < elements.Length)
      {
        Stopwatch stopwatch = new Stopwatch();

        Console.WriteLine("Highlighting {0} elements...", elements.Length);

        stopwatch.Start();

        // ハイライトしたイメージを取得する
        highlightedPartialGraph = (new Renderer(visualizer.RenderingContext)).GetElementImage(elements, true);

        stopwatch.Stop();

        Console.WriteLine("  Elaplsed: {0}", stopwatch.Elapsed);
        Console.WriteLine();
      }

      // 再描画
      viewingControl.Invalidate();
    }

    /// <summary>
    /// 指定されたツリーを見る
    /// </summary>
    public void ViewTree(Tree tree)
    {
      if (null != viewingHistory.Current && viewingHistory.Current.Tree == tree) return;

      ViewTree(tree, true, true, true, false);
    }

    /// <summary>
    /// グラフ上の任意の点に描画されているエントリを取得する
    /// </summary>
    private IGraphElement GetElementAt(Point point)
    {
      if (null == visualizer) return null;

      RenderingContext.RenderingParameter parameter = visualizer.RenderingContext.Parameter;
      PointF elementPoint = PointF.Empty;

      switch (viewingHistory.Current.Graph.Type)
      {
        case GraphType.Disc:
          // 現在の描画拡大縮小率を求める
          float ratio = (float)Math.Max(parameter.OutputWidth / (float)viewingControl.Width, parameter.OutputHeight / (float)viewingControl.Height);

          // 拡大縮小率を元に、元のサイズでの座標に変換する
          PointF imageCenter = new PointF(viewingControl.Width * 0.5f, viewingControl.Height * 0.5f);

          PointF discCenter = new PointF(parameter.OutputWidth * 0.5f, parameter.OutputHeight * 0.5f);

          elementPoint = new PointF(discCenter.X + (point.X - imageCenter.X) * ratio, discCenter.Y + (point.Y - imageCenter.Y) * ratio);

          break;

        default:
          return null;
      }

      // 特定の位置にあるエントリを取得する
      IGraphElement element = elementRetrievingRenderer.GetElementAt(elementPoint);
      
      if (null == element)
      {
        return null;
      }
      else
      {
        return element;
      }
    }

    private void RenderingParameterChanged()
    {
      SetRenderingParam(this.visualizer.RenderingContext.Parameter);

      if (null != this.viewingHistory.Current)
      {
        ViewHistory(this.viewingHistory.Current, false, true);

        // 描画済みの画像を参照させないよう、現在のヒストリ以外をクリアする
        viewingHistory.Clear(true);

        OnHistoryChanged(EventArgs.Empty);
      }
    }

    private void GraphingParameterChanged()
    {
      // 指定の方法で並び替える
      Comparison<Entry> comparer = Comparison.GetComparer(this.entryCompareMode, this.compareEntryType);

      if (this.entryComparer != comparer)
      {
        this.entryComparer = comparer;

        SetGraphingParam(visualizer.GraphingContext.Parameter);

        if (null != viewingHistory.Current)
        {
          ViewHistory(viewingHistory.Current, true, true);

          // 描画済みの画像を参照させないよう、ヒストリをクリアする
          viewingHistory.Clear(true);

          OnHistoryChanged(EventArgs.Empty);
        }
      }
    }

    protected virtual void OnEntryAtCursorChanged(ViewerMouseEventArgs e)
    {
      if (null != ElementAtCursorChanged) ElementAtCursorChanged(this, e);
    }

    protected virtual void OnEntryClicked(ViewerMouseEventArgs e)
    {
      if (null != ElementClicked) ElementClicked(this, e);
    }

    protected virtual void OnStatusChanged(ViewerStatusEventArgs e)
    {
      if (null != StatusChanged) StatusChanged(this, e);
    }

    protected virtual void OnGraphRootChanged(EventArgs e)
    {
      if (null != GraphRootChanged) GraphRootChanged(this, e);
    }

    protected virtual void OnHistoryChanged(EventArgs e)
    {
      if (null != HistoryChanged) HistoryChanged(this, e);
    }
  }
}
