#if WINDOWS

using System;
using System.Collections.Generic;
using System.IO;
using System.Runtime.InteropServices;

// this file is based on SantaMarta.Win32APIWrapper.dll

namespace Smdn.Applications.DirectoryAnalyzer.Util.Windows
{
  public static class AssociatedIcon
  {
    #region "internal types"

    public enum IconSize
    {
      Large,
      Small
    }

    #endregion

    /// <summary>
    /// パスで指定されたファイルまたはフォルダに関連づけられたアイコンの総数を取得します。
    /// </summary>
    /// <param name="path">対象のファイルまたはフォルダのパス</param>
    /// <param name="hInstance">HINSTANCE</param>
    /// <returns>含まれるアイコンの数。</returns>
    public static int GetIconCount(string path, IntPtr hInstance)
    {
      if (!File.Exists(path) && !Directory.Exists(path)) throw new FileNotFoundException("ファイルまたはディレクトリが見つかりませんでした。", path);

      if (IntPtr.Zero.Equals(hInstance)) throw new ArgumentNullException("hInstance");

      IntPtr result = PlatformApi.ExtractIcon(hInstance, path, -1);

      if (result.Equals(IntPtr.Zero))
      {
        return 0;
      }
      else
      {
        return result.ToInt32();
      }
    }

    /// <summary>
    /// パスで指定されたファイルまたはフォルダからアイコンを抜き出します。
    /// </summary>
    /// <param name="path">対象のファイルまたはフォルダのパス</param>
    /// <param name="hInstance">HINSTANCE</param>
    /// <returns>抜き出された<see cref="Icon"/>。</returns>
    public static Icon Extract(string path, IntPtr hInstance)
    {
      Icon[] icons = Extract(path, 0, 1, IconSize.Large, hInstance);

      if (0 < icons.Length)
      {
        return icons[0];
      }
      else
      {
        return null;
      }
    }

    /// <summary>
    /// パスで指定されたファイルまたはフォルダからアイコンを抜き出します。
    /// </summary>
    /// <param name="path">対象のファイルまたはフォルダのパス</param>
    /// <param name="iconSize">抜き出すアイコンのサイズ</param>
    /// <param name="hInstance">HINSTANCE</param>
    /// <returns>抜き出された<see cref="Icon"/>。</returns>
    public static Icon Extract(string path, IconSize iconSize, IntPtr hInstance)
    {
      Icon[] icons = Extract(path, 0, 1, iconSize, hInstance);

      if (0 < icons.Length)
      {
        return icons[0];
      }
      else
      {
        return null;
      }
    }

    /// <summary>
    /// パスで指定されたファイルまたはフォルダからアイコンを抜き出します。
    /// </summary>
    /// <param name="path">対象のファイルまたはフォルダのパス</param>
    /// <param name="index">アイコンのインデックス</param>
    /// <param name="hInstance">HINSTANCE</param>
    /// <returns>抜き出された<see cref="Icon"/>。</returns>
    public static Icon Extract(string path, int index, IntPtr hInstance)
    {
      Icon[] icons = Extract(path, index, 1, IconSize.Large, hInstance);

      if (0 < icons.Length)
      {
        return icons[0];
      }
      else
      {
        return null;
      }
    }

    /// <summary>
    /// パスで指定されたファイルまたはフォルダからアイコンを抜き出します。
    /// </summary>
    /// <param name="path">対象のファイルまたはフォルダのパス</param>
    /// <param name="index">アイコンのインデックス</param>
    /// <param name="iconSize">抜き出すアイコンのサイズ</param>
    /// <param name="hInstance">HINSTANCE</param>
    /// <returns>抜き出された<see cref="Icon"/>。</returns>
    public static Icon Extract(string path, int index, IconSize iconSize, IntPtr hInstance)
    {
      Icon[] icons = Extract(path, index, 1, iconSize, hInstance);

      if (0 < icons.Length)
      {
        return icons[0];
      }
      else
      {
        return null;
      }
    }

    /// <summary>
    /// パスで指定されたファイルまたはフォルダからアイコンを抜き出します。
    /// </summary>
    /// <param name="path">対象のファイルまたはフォルダのパス</param>
    /// <param name="index">アイコンのインデックス、または抜き出すアイコンの最初のインデックス</param>
    /// <param name="count">抜き出すアイコンの数</param>
    /// <param name="iconSize">抜き出すアイコンのサイズ</param>
    /// <param name="hInstance">HINSTANCE</param>
    /// <returns>抜き出された<see cref="Icon"/>の配列。</returns>
    public static Icon[] Extract(string path, int index, int count, IconSize iconSize, IntPtr hInstance)
    {
      if (!File.Exists(path) && !Directory.Exists(path)) throw new FileNotFoundException("ファイルまたはディレクトリが見つかりませんでした。", path);

      if (index < 0) throw new ArgumentOutOfRangeException("index", "indexは0以上でなければなりません。");

      if (count <= 0) throw new ArgumentOutOfRangeException("count", "countは1以上でなければなりません。");

      if (IntPtr.Zero.Equals(hInstance)) throw new ArgumentNullException("hInstance");

      // アイコン一覧
      IntPtr[] hIcons = new IntPtr[count];

      for (int i = 0; i < count; i++)
      {
        hIcons[i] = IntPtr.Zero;
      }

      // 取得する
      int iconCount = 0;

      if (iconSize == IconSize.Large)
      {
        // 大きいアイコンを取得
        iconCount = PlatformApi.ExtractIconEx(path, index, hIcons, null, count);
      }
      else if (iconSize == IconSize.Small)
      {
        // 小さいアイコンを取得
        iconCount = PlatformApi.ExtractIconEx(path, index, null, hIcons, count);
      }

      List<Icon> icons = new List<Icon>();

      if (0 < iconCount)
      {
        // 取得出来た場合、ハンドルからアイコンを作成
        try
        {
          for (int i = 0; i < iconCount; i++)
          {
            if (IntPtr.Zero.Equals(hIcons[i])) continue;

            Icon icon = Icon.FromHIcon(hIcons[i]);

            if (icon != null) icons.Add(icon);
          }
        }
        finally
        {
          for (int i = 0; i < iconCount; i++)
          {
            if (IntPtr.Zero.Equals(hIcons[i])) continue;

            // ハンドルを解放
            PlatformApi.DestroyIcon(hIcons[i]);
          }
        }
      }
      else
      {
        // 取得できなかった場合、関連付けられたアイコンを取得
        Icon icon = ExtractAssociatedIcon(path, hInstance);

        if (icon != null) icons.Add(icon);
      }

      return icons.ToArray();
    }

    /// <summary>
    /// パスで指定されたファイルまたはフォルダからアイコンをすべて抜き出します。
    /// </summary>
    /// <param name="path">対象のファイルまたはフォルダのパス</param>
    /// <param name="hInstance">HINSTANCE</param>
    /// <returns>抜き出された<see cref="Icon"/>の配列。</returns>
    public static Icon[] ExtractAll(string path, IntPtr hInstance)
    {
      return ExtractAll(path, IconSize.Large, hInstance);
    }

    /// <summary>
    /// パスで指定されたファイルまたはフォルダからアイコンをすべて抜き出します。
    /// </summary>
    /// <param name="path">対象のファイルまたはフォルダのパス</param>
    /// <param name="iconSize">抜き出すアイコンのサイズ</param>
    /// <param name="hInstance">HINSTANCE</param>
    /// <returns>抜き出された<see cref="Icon"/>の配列。</returns>
    public static Icon[] ExtractAll(string path, IconSize iconSize, IntPtr hInstance)
    {
      // 取得可能なアイコン数
      int iconCount = GetIconCount(path, hInstance);

      if (iconCount <= 0)
      {
        Icon icon = ExtractAssociatedIcon(path, hInstance);

        if (icon == null)
        {
          return new Icon[] { };
        }
        else
        {
          return new Icon[] { icon };
        }
      }
      else
      {
        return Extract(path, 0, iconCount, iconSize, hInstance);
      }
    }

    /// <summary>
    /// パスで指定されたファイルまたはフォルダに関連付けられたアイコンを取得する
    /// </summary>
    /// <param name="path">対象のファイルまたはフォルダのパス</param>
    /// <param name="hInstance">HINSTANCE</param>
    /// <returns>抜き出された<see cref="Icon"/>。</returns>
    public static Icon ExtractAssociatedIcon(string path, IntPtr hInstance)
    {
      return ExtractAssociatedIcon(path, 0, hInstance);
    }

    /// <summary>
    /// パスで指定されたファイルまたはフォルダに関連付けられたアイコンを取得する
    /// </summary>
    /// <param name="path">対象のファイルまたはフォルダのパス</param>
    /// <param name="index">アイコンのインデックス、または抜き出すアイコンの最初のインデックス</param>
    /// <param name="hInstance">HINSTANCE</param>
    /// <returns>抜き出された<see cref="Icon"/>。</returns>
    public static Icon ExtractAssociatedIcon(string path, int index, IntPtr hInstance)
    {
      if (!File.Exists(path) && !Directory.Exists(path)) throw new FileNotFoundException("ファイルまたはディレクトリが見つかりませんでした。", path);

      if (index < 0) throw new ArgumentOutOfRangeException("index", "indexは0以上でなければなりません。");

      if (IntPtr.Zero.Equals(hInstance)) throw new ArgumentNullException("hInstance");

      // 関連付けられたアイコンを取得
      IntPtr hIcon = PlatformApi.ExtractAssociatedIcon(hInstance, path, ref index);

      if (IntPtr.Zero.Equals(hIcon))
      {
        return null;
      }

      // ハンドルからアイコンを作成
      Icon icon = null;

      try
      {
        icon = Icon.FromHIcon(hIcon);
      }
      finally
      {
        // ハンドルを解放
        PlatformApi.DestroyIcon(hIcon);
      }

      return icon;
    }
  }
}
#endif
