#if WINDOWS

using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.IO;
using System.Runtime.InteropServices;

namespace Smdn.Applications.DirectoryAnalyzer.Util.Windows
{
  public class Icon
  {
    #region "internal types"

    [StructLayout(LayoutKind.Explicit)]
    public struct Argb
    {
      [FieldOffset(0)]
      public uint Value;

      [FieldOffset(0)]
      public byte B;
      [FieldOffset(1)]
      public byte G;
      [FieldOffset(2)]
      public byte R;
      [FieldOffset(3)]
      public byte A;
    }

    public enum ColorDepth : int
    {
      Depth1Bit = 1,
      Depth4Bit = 4,
      Depth8Bit = 8,
      Depth16Bit = 16,
      Depth24Bit = 24,
      Depth32Bit = 32,
    }

    #endregion

    #region "fields and properties"

    private int width;

    public int Width
    {
      get
      {
        return width;
      }
    }

    private int height;

    public int Height
    {
      get
      {
        return height;
      }
    }

    private byte[,] mask = null;

    public byte[,] Mask
    {
      get
      {
        return mask;
      }
    }

    private Argb[,] image = null;

    public Argb[,] Image
    {
      get
      {
        return image;
      }
    }

    private ColorDepth originalColorDepth;

    public ColorDepth OriginalColorDepth
    {
      get
      {
        return originalColorDepth;
      }
    }

    #endregion

    #region "constructors and destructors"

    protected internal Icon(Argb[,] image, byte[,] mask)
      : this(image, mask, ColorDepth.Depth32Bit)
    {
    }

    protected internal Icon(Argb[,] image, byte[,] mask, ColorDepth originalColorDepth)
    {
      this.image = image;
      this.mask = mask;

      width = image.GetLength(0);
      height = image.GetLength(1);

      if (width != mask.GetLength(0)) throw new ArgumentException("イメージとマスクの幅が一致しません。");
      if (height != mask.GetLength(1)) throw new ArgumentException("イメージとマスクの高さが一致しません。");

      this.originalColorDepth = originalColorDepth;
    }

    public Icon(int width, int height)
    {
      this.width = width;
      this.height = height;

      image = Array.CreateInstance(typeof(Argb), width, height) as Argb[,];
      mask = Array.CreateInstance(typeof(byte), width, height) as byte[,];

      for (int y = 0; y < height; y++)
      {
        for (int x = 0; x < width; x++)
        {
          image[x, y].Value = 0;
          mask[x, y] = 0xff;
        }
      }

      this.originalColorDepth = ColorDepth.Depth32Bit;
    }

    public static Icon FromRawImage(Argb[,] image, byte[,] mask)
    {
      if (image == null) throw new ArgumentNullException("image");
      if (mask == null) throw new ArgumentNullException("mask");

      return new Icon(image, mask);
    }

    public static Icon FromHIcon(IntPtr hIcon)
    {
      if (IntPtr.Zero.Equals(hIcon)) throw new ArgumentNullException("hIconがNULLです。");

      try
      {
        return FromHIconDefaultImpl(hIcon);
      }
      catch (PlatformApiException)
      {
        return FromHIconAlternateImpl(hIcon);
      }
    }

    private static Icon FromHIconDefaultImpl(IntPtr hIcon)
    {
      ICONINFO iconInfo = new ICONINFO();

      if (!PlatformApi.GetIconInfo(hIcon, out iconInfo)) throw new ArgumentException("アイコンを取得出来ませんでした。");

      IntPtr hDC = IntPtr.Zero;

      try
      {
        hDC = PlatformApi.CreateCompatibleDC(IntPtr.Zero);

        if (IntPtr.Zero.Equals(hDC))
        {
          throw new PlatformApiException();
        }

        // イメージを読み込む
        BITMAPINFO_32BPP bmi = new BITMAPINFO_32BPP();

        bmi.bmiHeader.biSize = (uint)Marshal.SizeOf(bmi.bmiHeader);

        if (0 == PlatformApi.GetDIBits(hDC, iconInfo.hbmColor, 0, 0, IntPtr.Zero, ref bmi, PlatformApi.DIB_RGB_COLORS))
        {
          throw new PlatformApiException();
        }

        ColorDepth colorDepth = ColorDepth.Depth32Bit;

        switch (bmi.bmiHeader.biBitCount)
        {
          case 1:
            colorDepth = ColorDepth.Depth1Bit;
            break;

          case 4:
            colorDepth = ColorDepth.Depth4Bit;
            break;

          case 8:
            colorDepth = ColorDepth.Depth8Bit;
            break;

          case 16:
            colorDepth = ColorDepth.Depth16Bit;
            break;

          case 24:
            colorDepth = ColorDepth.Depth24Bit;
            break;

          case 32:
            colorDepth = ColorDepth.Depth32Bit;
            break;

          default:
            throw new NotSupportedException("サポートしていない色深度");
        }

        int w = bmi.bmiHeader.biWidth;
        int h = bmi.bmiHeader.biHeight;

        Argb[,] image = Array.CreateInstance(typeof(Argb), w, h) as Argb[,];

        IntPtr bits = IntPtr.Zero;

        try
        {
          bits = Marshal.AllocHGlobal((int)bmi.bmiHeader.biSizeImage);

          int stride = (((w * bmi.bmiHeader.biPlanes * bmi.bmiHeader.biBitCount) + 31) & ~31) / 8;

          // TODO: GetDIBits fails on Windows 7 (64bit)
          if (h != PlatformApi.GetDIBits(hDC, iconInfo.hbmColor, 0, (uint)h, bits, ref bmi, PlatformApi.DIB_RGB_COLORS))
          {
            throw new PlatformApiException();
          }

          for (int y = 0; y < h; y++)
          {
            int offset = (h - 1 - y) * stride;

            for (int x = 0; x < w; x++)
            {
              image[x, y].Value = (uint)Marshal.ReadInt32(bits, offset);

              offset += 4;
            }
          }
        }
        finally
        {
          if (!IntPtr.Zero.Equals(bits))
          {
            Marshal.FreeHGlobal(bits);

            bits = IntPtr.Zero;
          }
        }

        // マスクを読み込む
        byte[,] mask = Array.CreateInstance(typeof(byte), w, h) as byte[,];

        if (0 != bmi.bmiHeader.biClrUsed)
        {
          BITMAPINFO_1BPP bmi1bit = new BITMAPINFO_1BPP();

          bmi1bit.bmiHeader.biSize = (uint)Marshal.SizeOf(bmi1bit.bmiHeader);

          if (0 == PlatformApi.GetDIBits(hDC, iconInfo.hbmMask, 0, 0, IntPtr.Zero, ref bmi1bit, PlatformApi.DIB_PAL_COLORS))
          {
            throw new PlatformApiException();
          }

          try
          {
            bits = Marshal.AllocHGlobal((int)bmi1bit.bmiHeader.biSizeImage);

            int stride = (((w * bmi1bit.bmiHeader.biPlanes * bmi1bit.bmiHeader.biBitCount) + 31) & ~31) / 8;

            if (h != PlatformApi.GetDIBits(hDC, iconInfo.hbmMask, 0, (uint)h, bits, ref bmi1bit, PlatformApi.DIB_RGB_COLORS))
            {
              throw new PlatformApiException();
            }

            for (int y = 0; y < h; y++)
            {
              int offset = (h - 1 - y) * stride;

              for (int x = 0; x < w; x++)
              {
                mask[x, y] = ((byte)((Marshal.ReadByte(bits, offset + x / 8) >> (7 - x % 8)) & 0x1) == 0) ? (byte)0xff : (byte)0x00;
              }
            }
          }
          finally
          {
            if (!IntPtr.Zero.Equals(bits))
            {
              Marshal.FreeHGlobal(bits);

              bits = IntPtr.Zero;
            }
          }
        }
        else
        {
          for (int y = 0; y < h; y++)
          {
            for (int x = 0; x < w; x++)
            {
              mask[x, y] = (byte)0xff;
            }
          }
        }

        return new Icon(image, mask, colorDepth);

      }
      finally
      {
        if (!IntPtr.Zero.Equals(hDC)) PlatformApi.DeleteDC(hDC);
      }
    }

    private static Icon FromHIconAlternateImpl(IntPtr hIcon)
    {
      using (System.Drawing.Icon icon = System.Drawing.Icon.FromHandle(hIcon))
      {
        using (System.Drawing.Bitmap bitmap = icon.ToBitmap())
        {
          BitmapData bitmapData = null;

          try
          {
            bitmapData = bitmap.LockBits(new Rectangle(0, 0, bitmap.Width, bitmap.Height), ImageLockMode.ReadOnly, PixelFormat.Format32bppArgb);

            Argb[,] image = Array.CreateInstance(typeof(Argb), bitmap.Width, bitmap.Height) as Argb[,];
            byte[,] mask = Array.CreateInstance(typeof(byte), bitmap.Width, bitmap.Height) as byte[,];

            // 色部分の出力
            for (int y = 0; y < bitmap.Height; y++)
            {
              int imageIndex = bitmapData.Stride * y;

              for (int x = 0; x < bitmap.Width; x++)
              {
                image[x, y].B = Marshal.ReadByte(bitmapData.Scan0, imageIndex);
                image[x, y].G = Marshal.ReadByte(bitmapData.Scan0, imageIndex + 1);
                image[x, y].R = Marshal.ReadByte(bitmapData.Scan0, imageIndex + 2);
                image[x, y].A = Marshal.ReadByte(bitmapData.Scan0, imageIndex + 3);

                mask[x, y] = 0xff;

                imageIndex += 4;
              }
            }

            return new Icon(image, mask);
          }
          finally
          {
            if (bitmapData != null)
            {
              bitmap.UnlockBits(bitmapData);
            }
          }
        }
      }
    }

    public static Icon FromIconImage(ICONIMAGE iconImage)
    {
      int width = iconImage.icHeader.biWidth;
      int height = iconImage.icHeader.biHeight / 2;

      // 色深度
      int bitCount = iconImage.icHeader.biBitCount;
      ColorDepth colorDepth = ColorDepth.Depth32Bit;

      switch (bitCount)
      {
        case 1:
          colorDepth = ColorDepth.Depth1Bit;
          break;

        case 4:
          colorDepth = ColorDepth.Depth4Bit;
          break;

        case 8:
          colorDepth = ColorDepth.Depth8Bit;
          break;

        case 16:
          colorDepth = ColorDepth.Depth16Bit;
          break;

        case 24:
          colorDepth = ColorDepth.Depth24Bit;
          break;

        case 32:
          colorDepth = ColorDepth.Depth32Bit;
          break;

        default:
          throw new NotSupportedException("サポートしていない色深度");
      }

      // 1ラインあたりのバイト数
      int bytesPerLine = (((iconImage.icHeader.biWidth * iconImage.icHeader.biPlanes * bitCount) + 31) & ~31) / 8;

      // イメージを解釈
      Argb[,] image = Array.CreateInstance(typeof(Argb), width, height) as Argb[,];

      for (int y = 0; y < height; y++)
      {
        int offset = bytesPerLine * (height - 1 - y);

        for (int x = 0; x < width; x++)
        {
          image[x, y] = new Argb();

          if (bitCount <= 8)
          {
            // indexed color
            RGBQUAD color = new RGBQUAD();

            if (bitCount == 1)
            {
              byte paletteIndex = (byte)((iconImage.icXor[offset + x / 8] >> (7 - x % 8)) & 0x1);

              color = iconImage.icColors[paletteIndex];
            }
            else if (bitCount == 4)
            {
              byte paletteIndex = (byte)((iconImage.icXor[offset + x / 2] >> (4 * (1 - x % 2))) & 0xf);

              color = iconImage.icColors[paletteIndex];
            }
            else if (bitCount == 8)
            {
              color = iconImage.icColors[iconImage.icXor[offset + x]];
            }
            image[x, y].B = color.rgbBlue;
            image[x, y].G = color.rgbGreen;
            image[x, y].R = color.rgbRed;
            image[x, y].A = 0xff;
          }
          else
          {

            if (bitCount == 16)
            {
              throw new NotImplementedException("未実装");
            }
            else if (bitCount == 24)
            {
              int index = offset + x * 3;

              image[x, y].B = iconImage.icXor[index];
              image[x, y].G = iconImage.icXor[index + 1];
              image[x, y].R = iconImage.icXor[index + 2];
              image[x, y].A = 0xff;

            }
            else if (bitCount == 32)
            {
              int index = offset + x * 4;

              image[x, y].B = iconImage.icXor[index];
              image[x, y].G = iconImage.icXor[index + 1];
              image[x, y].R = iconImage.icXor[index + 2];
              image[x, y].A = iconImage.icXor[index + 3];
            }
          }
        }
      }

      // マスクを解釈
      byte[,] mask = Array.CreateInstance(typeof(byte), width, height) as byte[,];

      bytesPerLine = ((iconImage.icHeader.biWidth + 31) & ~31) / 8;

      for (int y = 0; y < height; y++)
      {
        int offset = bytesPerLine * (height - 1 - y);

        for (int x = 0; x < width; x++)
        {
          mask[x, y] = ((byte)(((iconImage.icAnd[offset + x / 8]) >> (7 - x % 8)) & 0x1) == 0) ? (byte)0xff : (byte)0x00;
        }
      }

      return new Icon(image, mask, colorDepth);
    }

    public static Icon[] FromIconStream(Stream stream)
    {
      if (stream == null) throw new ArgumentNullException("stream");

      BinaryReader reader = new BinaryReader(stream);

      // アイコンのデータ
      ICONDIR icondir = new ICONDIR();

      icondir.idReserved = reader.ReadUInt16();
      icondir.idType = reader.ReadUInt16();
      icondir.idCount = reader.ReadUInt16();

      // フォーマットが違う
      if (icondir.idType != 1)
      {
        // イメージとしてストリームを読み込み直す
        stream.Seek(0, SeekOrigin.Begin);

        return new Icon[] { FromImageStream(stream) };
      }

      // アイコン数が異常
      if (icondir.idCount <= 0 || 0x400 < icondir.idCount) throw new ArgumentException("含まれるアイコンの数が不正です。");

      //icondir.idEntries = Array.CreateInstance(typeof(ICONDIRENTRY), icondir.idCount) as ICONDIRENTRY[];
      //icondir.idImages = Array.CreateInstance(typeof(ICONIMAGE), icondir.idCount) as ICONIMAGE[];

      // 読み込みを開始する
      List<Icon> icons = new List<Icon>();
      long position = -1;

      for (int index = 0; index < icondir.idCount; index++)
      {
        if (0 <= position) reader.BaseStream.Seek(position, SeekOrigin.Begin);

        ICONDIRENTRY entry = new ICONDIRENTRY();

        entry.bWidth = reader.ReadByte();
        entry.bHeight = reader.ReadByte();
        entry.bColorCount = reader.ReadByte();
        entry.bReserved = reader.ReadByte();
        entry.wPlanes = reader.ReadUInt16();
        entry.wBitCount = reader.ReadUInt16();
        entry.dwBytesInRes = reader.ReadUInt32();
        entry.dwImageOffset = reader.ReadUInt32();

        //icondir.idEntries[index] = entry;

        position = reader.BaseStream.Position;

        reader.BaseStream.Seek(entry.dwImageOffset, SeekOrigin.Begin);

        MemoryStream resourceStream = new MemoryStream(reader.ReadBytes((int)entry.dwBytesInRes));
        BinaryReader resourceReader = new BinaryReader(resourceStream);

        BITMAPINFOHEADER bih = new BITMAPINFOHEADER();

        bih.biSize = resourceReader.ReadUInt32();
        bih.biWidth = resourceReader.ReadInt32();
        bih.biHeight = resourceReader.ReadInt32();
        bih.biPlanes = resourceReader.ReadUInt16();
        bih.biBitCount = resourceReader.ReadUInt16();
        bih.biCompression = resourceReader.ReadUInt32();
        bih.biSizeImage = resourceReader.ReadUInt32();
        bih.biXPelsPerMeter = resourceReader.ReadInt32();
        bih.biYPelsPerMeter = resourceReader.ReadInt32();
        bih.biClrUsed = resourceReader.ReadUInt32();
        bih.biClrImportant = resourceReader.ReadUInt32();

        if (bih.biCompression != 0) throw new NotSupportedException("圧縮されたビットマップには対応していません。");

        ICONIMAGE image = new ICONIMAGE();

        image.icHeader = bih;

        if (bih.biClrUsed != 0)
        {
          image.icColors = Array.CreateInstance(typeof(RGBQUAD), bih.biClrUsed) as RGBQUAD[];
        }
        else
        {
          if (bih.biBitCount < 16)
          {
            image.icColors = Array.CreateInstance(typeof(RGBQUAD), (int)Math.Pow(2.0, bih.biBitCount)) as RGBQUAD[];
          }
          else
          {
            image.icColors = new RGBQUAD[] { };
          }
        }

        for (int paletteIndex = 0; paletteIndex < image.icColors.Length; paletteIndex++)
        {
          RGBQUAD color = new RGBQUAD();

          color.rgbBlue = resourceReader.ReadByte();
          color.rgbGreen = resourceReader.ReadByte();
          color.rgbRed = resourceReader.ReadByte();
          color.rgbReserved = resourceReader.ReadByte();

          image.icColors[paletteIndex] = color;
        }

        int iconHeight = bih.biHeight / 2;
        int bytesPerLine = (((bih.biWidth * bih.biPlanes * bih.biBitCount) + 31) & ~31) / 8;

        image.icXor = resourceReader.ReadBytes(bytesPerLine * iconHeight);
        image.icAnd = resourceReader.ReadBytes(bytesPerLine * iconHeight);

        //icondir.idImages[index] = image;

        icons.Add(FromIconImage(image));
      }

      return icons.ToArray();
    }

    internal static Icon FromImageStream(Stream stream)
    {
      Bitmap b;

      try
      {
        b = new Bitmap(stream);

        BitmapData bitmapData = null;

        Icon icon = null;

        try
        {
          bitmapData = b.LockBits(new Rectangle(0, 0, b.Width, b.Height), ImageLockMode.ReadOnly, PixelFormat.Format32bppArgb);

          IntPtr scan0 = bitmapData.Scan0;
          int stride = bitmapData.Stride;

          Argb[,] image = Array.CreateInstance(typeof(Argb), b.Width, b.Height) as Argb[,];
          byte[,] mask = Array.CreateInstance(typeof(byte), b.Width, b.Height) as byte[,];

          // 色部分の出力
          for (int y = 0; y < b.Height; y++)
          {
            for (int x = 0; x < b.Width; x++)
            {
              int imageIndex = stride * y + x * 4;

              image[x, y].B = Marshal.ReadByte(scan0, imageIndex);
              image[x, y].G = Marshal.ReadByte(scan0, imageIndex + 1);
              image[x, y].R = Marshal.ReadByte(scan0, imageIndex + 2);
              image[x, y].A = Marshal.ReadByte(scan0, imageIndex + 3);

              mask[x, y] = 0xff;
            }
          }

          icon = new Icon(image, mask);
        }
        finally
        {
          if (bitmapData != null)
          {
            b.UnlockBits(bitmapData);
          }
        }

        return icon;
      }
      catch
      {
        throw;
      }
    }

    #endregion
  }
}
#endif
