using System;
using System.Diagnostics;

namespace Smdn.Applications.DirectoryAnalyzer
{
  using Analysis;
  using FileSystem;
  using Format;
  using Graphing;
  using Render;

  public class Visualizer
  {
    private delegate ITaskContext PrepareTaskContext(bool resetParameter);
    private delegate void TaskStart(ITaskContext context);

    private TaskStart[] tasks;
    private PrepareTaskContext[] taskContextPreparer;

    private bool isCompleted = false;
    private int taskTotal = 0;
    private int taskCompleted = 0;
    private TaskContext.TaskStatus status = null;
    private AnalysisContext analysisContext = null;
    private GraphingContext graphingContext = null;
    private RenderingContext renderingContext = null;
    private FormattingContext formattingContext = null;

    public bool IsCompleted
    {
      get
      {
        return this.isCompleted;
      }
    }

    public TaskContext.TaskStatus Status
    {
      get
      {
        return status;
      }
    }

    public AnalysisContext AnalysisContext
    {
      get
      {
        return this.analysisContext;
      }
    }

    public GraphingContext GraphingContext
    {
      get
      {
        return this.graphingContext;
      }
    }

    public RenderingContext RenderingContext
    {
      get
      {
        return this.renderingContext;
      }
    }

    public FormattingContext FormattingContext
    {
      get
      {
        return this.formattingContext;
      }
    }

    public int TaskTotal
    {
      get
      {
        return this.taskTotal;
      }
    }

    public int TaskCompleted
    {
      get
      {
        return this.taskCompleted;
      }
    }

    public Visualizer(Directory directory)
    {
      if (null == directory) throw new ArgumentNullException("directory");

      this.isCompleted = false;

      this.analysisContext = new AnalysisContext(directory);

      this.graphingContext = new GraphingContext();

      this.renderingContext = new RenderingContext();

      this.formattingContext = new FormattingContext();

      InitializeTasks();
    }

    private void InitializeTasks()
    {
      // タスク
      tasks = new TaskStart[]
      {
        delegate(ITaskContext context)
        {
          AnalysisContext analysisContext = context as AnalysisContext;

          Console.Write("Analyzing directory... ");
          ConsoleEx.PreserveLoction();

          // 解析
          Analyzer analyzer = new Analyzer(analysisContext);

          this.status = analysisContext.Status;

          analyzer.Analyze();

          this.status = null;

          uint allEntryCount = analysisContext.Result.Tree.GetAllEntryCount();

          ConsoleEx.RevertLocation();
          ConsoleEx.WriteProgress((int)allEntryCount, (int)allEntryCount);
          Console.WriteLine();
          Console.WriteLine("Entries in this directory: {0}", allEntryCount);
          Console.WriteLine("Deepest directory depth: {0}", analysisContext.Result.Tree.Depth);
          if (0 <= analysisContext.Result.ExcludedDirectoryCount)
          {
            Console.WriteLine("Excluded directories: {0}", analysisContext.Result.ExcludedDirectoryCount);
          }
        },

        delegate(ITaskContext context)
        {
          GraphingContext graphingContext = context as GraphingContext;

          // サイズを計算(ツリー全体のエントリのサイズを一気に計算してキャッシュする)
          Console.Write("Calculating directory sizes(It might take a while)... ");
          ConsoleEx.PreserveLoction();

          context.Status.StatusText = "全体のサイズを計算中(少し時間がかかります)";

          this.status = graphingContext.Status;

          Size size = graphingContext.Parameter.Tree.RootDirectory.Size;

          Console.WriteLine();
          Console.WriteLine("Directory size: {0}", size);

          Console.Write("Generating graph... ");
          ConsoleEx.PreserveLoction();

          // グラフを作成
          Grapher grapher = new Grapher(graphingContext);

          grapher.Create();

          this.status = null;

          ConsoleEx.RevertLocation();
          ConsoleEx.WriteProgress((int)graphingContext.Status.Processed, (int)graphingContext.Status.Total);
          Console.WriteLine();
        },

        delegate(ITaskContext context)
        {
          RenderingContext renderingContext = context as RenderingContext;

          Console.Write("Rendering graph... ");
          ConsoleEx.PreserveLoction();

          // グラフをレンダリング
          Renderer renderer = new Renderer(renderingContext);

          this.status = renderingContext.Status;

          renderer.Render();

          this.status = null;

          ConsoleEx.RevertLocation();
          ConsoleEx.WriteProgress((int)renderingContext.Status.Processed, (int)renderingContext.Status.Total);
          Console.WriteLine();
        },

        delegate(ITaskContext context)
        {
          FormattingContext formattingContext = context as FormattingContext;

          Console.Write("Formatting graph... ");
          ConsoleEx.PreserveLoction();

          // グラフを整形
          Formatter formatter = new Formatter(formattingContext);

          this.status = formattingContext.Status;

          formatter.Format();

          this.status = null;

          ConsoleEx.RevertLocation();
          ConsoleEx.WriteProgress((int)formattingContext.Status.Processed, (int)formattingContext.Status.Total);
          Console.WriteLine();
        },
      };

      // 各タスクに引き渡すコンテキストを用意する処理
      taskContextPreparer = new PrepareTaskContext[]
      {
        delegate(bool reset)
        {
          return this.analysisContext;
        },

        delegate(bool reset)
        {
          if (reset)
          {
            this.graphingContext.Parameter.Tree = this.analysisContext.Result.Tree;
          }

          return this.graphingContext;
        },

        delegate(bool reset)
        {
          if (reset)
          {
            this.renderingContext.Parameter.Graph = this.graphingContext.Result.Graph;
          }
        
          return this.renderingContext;
        },

        delegate(bool reset)
        {
          if (reset)
          {
            this.formattingContext.Parameter.Graph = this.graphingContext.Result.Graph;
            this.formattingContext.Parameter.FormattingElements = this.graphingContext.Result.Graph.GetAllElements();
          }
        
          return this.formattingContext;
        },
      };
    }

    /// <summary>
    /// ディレクトリの解析、グラフの作成、グラフの描画を行う
    /// </summary>
    public void Visualize()
    {
      ExecuteTask(new int[] { 0, 1, 2 }, true);
    }

    /// <summary>
    /// ディレクトリの解析、グラフの作成、グラフの整形を行う
    /// </summary>
    public void Format()
    {
      ExecuteTask(new int[] { 0, 1, 3 }, true);
    }

    /// <summary>
    /// グラフの作成を行う
    /// </summary>
    public void Graph(bool resetParameter)
    {
      ExecuteTask(new int[] { 1 }, resetParameter);
    }

    /// <summary>
    /// グラフの描画を行う
    /// </summary>
    public void Render(bool resetParameter)
    {
      ExecuteTask(new int[] { 2 }, resetParameter);
    }

    /// <summary>
    /// グラフの整形を行う
    /// </summary>
    public void Format(bool resetParameter)
    {
      ExecuteTask(new int[] { 3 }, resetParameter);
    }

    /// <summary>
    /// 指定されたインデックスのタスクを実行する
    /// </summary>
    private void ExecuteTask(int[] executeTaskIndices, bool resetParameter)
    {
      this.isCompleted = false;
      this.taskTotal = executeTaskIndices.Length;
      this.taskCompleted = 0;

      // 各タスクを実行
      Stopwatch stopwatch = new Stopwatch();

      for (uint task = 0; task < executeTaskIndices.Length; task++)
      {
        stopwatch.Start();

        try
        {
          int taskIndex = executeTaskIndices[task];

          // タスクに引き渡すコンテキストを準備
          ITaskContext context = taskContextPreparer[taskIndex](resetParameter);

          // タスクを起動
          tasks[taskIndex](context);

          taskCompleted++;
        }
        catch (System.Threading.ThreadAbortException)
        {
          Console.Error.WriteLine("Abort requested");

          throw;
        }
        catch (Exception ex)
        {
          Console.Error.WriteLine(ex);

          throw ex;
        }

        // 経過時間を出力
        stopwatch.Stop();

        Console.WriteLine("  Elapsed: {0}", stopwatch.Elapsed);
        Console.WriteLine();

        stopwatch.Reset();
      }

      // 完了
      this.isCompleted = true;

      return;
    }
  }
}
